''' Supports annotation browsing functionality.
'''

import numpy as np

from kivy.app import App
from kivy.logger import Logger
from kivy.properties import (
    NumericProperty,
    StringProperty,
    BooleanProperty,
    ListProperty,
    DictProperty,
)
from kivy.uix.label import Label
from kivy.uix.boxlayout import BoxLayout
from kivy.metrics import dp
from kivy.lang import Builder
from kivy.core.window import Window

from jocular.hoverable import HoverBehavior
from jocular.metrics import Metrics
from jocular.component import Component
from jocular.uranography import make_tile

Builder.load_string(
'''

#:import np numpy

<PropVal>:
    canvas.before:
        Color:
            rgba: .3, .3, .3, .9
        Rectangle:
            pos: self.pos
            size: self.size
    size_hint: 1, None
    height: dp(20)
    Label:
        size_hint: None, 1
        markup: True
        halign: 'left'
        text_size: self.size
        color: app.lowlight_color
        font_size: app.info_font_size
        background_color: .6, 0, .6, 0
        width: dp(100)
        padding: dp(5), dp(1)
        text: '[b]{:}[/b]'.format(root.param)
    Label:
        size_hint: None, 1
        markup: True
        halign: 'left'
        text_size: self.size
        color: app.lowlight_color
        font_size: app.info_font_size
        background_color: .6, 0, .6, 0
        width: dp(250)
        text: '{:}'.format(root.value)

<TitleBar>:
    canvas.before:
        Color:
            rgba: .4, .4, .4, .3
        Rectangle:
            pos: self.pos
            size: self.size
    size_hint: 1, None
    height: dp(30)
    padding: dp(5), dp(2)
    Label:
        size_hint: None, 1
        markup: True
        halign: 'left'
        shorten: True
        shorten_from: 'right'
        valign: 'middle'
        text_size: self.size
        color: root.title_color
        font_size: str(int(app.info_font_size[:-2]) + 4) + 'sp'
        width: dp(350)
        text: '[b]{:}[/b]'.format(root.Name)

<DSOPopup>:
    size_hint: None, 1
    width: dp(350)
    orientation: 'vertical'
    padding: dp(5), dp(2)

<Annotation>:
    size_hint: None, None
    size: self.texture_size
    font_size: '15sp'
    text_size: None, None  # forces size to be that of text
    color: (self.label_color if self.pinned else self.marker_color) + [self.visible]
    pos: (self.lab_x - self.width / 2, self.lab_y - self.height / 2) if self.center and not self.pinned else (self.lab_x, self.lab_y)

<AnnotCluster>:
    canvas:
        Color:
            rgb: 1, 0, 0
            a: 1 if self.visible else 0
        Line:
            circle: self.sx, self.sy, self.radius
            dash_offset: 16
            dash_length: 4
            width: 1

<AnnotQuasar>:
    canvas:
        Color:
            rgb: 1, 0, 0
            a: 1 if self.visible else 0
        Line:
            points: [self.sx - self.gap - self.length, self.sy, self.sx - self.gap, self.sy]
            width: 1
        Line:
            points: [self.sx, self.sy - self.gap - self.length, self.sx, self.sy - self.gap]
            width: 1

<AnnotFOV>:
    canvas:
        Color:
            rgb: .8, .8, .8
            a: 1 if self.visible else 0
        Line:
            points: [self.sx - self.length/2, self.sy, self.sx + self.length/2, self.sy]
            width: 1
    color: .8, .8, .8, 1

'''
)


class Annotation(Label, HoverBehavior):

    px = NumericProperty(0)  # image pixel centre
    py = NumericProperty(0)
    sx = NumericProperty(0)  # screen pixel centre
    sy = NumericProperty(0)

    ot = StringProperty('')
    ang = NumericProperty(45)
    visible = BooleanProperty(False)  # in eyepiece and within mag limit
    display = BooleanProperty(False)  # within mag limit
    pinned = BooleanProperty(False)
    marker_color = ListProperty([0.5, 0.3, 0.3])
    label_color = ListProperty([0.5, 0.5, 0.5])
    infoline = StringProperty('')
    info = DictProperty({})
    radius = NumericProperty(0)
    bx = NumericProperty(0)  #  pixel on boundary
    by = NumericProperty(0)
    lab_x = NumericProperty(0)
    lab_y = NumericProperty(0)
    center = BooleanProperty(True)
    mag = NumericProperty(0)

    def on_enter(self, *args):
        if self.visible:
            self.display_DSOInfo()

    def on_leave(self, *args):
        if self.visible:
            self.close_DSOInfo()

    def on_touch_down(self, touch):
        if self.visible:
            if self.collide_point(*touch.pos):
                if not self.pinned:
                    self.pin()
                else:
                    self.unpin()
                return True
        return False

    def pin(self, dt=None):
        if not self.pinned:
            lab = self.info['Name']
            self.font_size = '15sp'
            # for quasars, label with mag + redshift
            if self.ot == 'QS':
                lab = '{:4.1f}'.format(self.info['Mag'])
                z = self.info.get('z', '?')
                if not np.isnan(z):
                    lab += ' z {:3.1f}'.format(z)
                self.font_size = '14sp'
            # for members, use member number
            elif self.ot == 'ME':
                lab = self.info['Mem']
            self.text = str(lab)
            self.pinned = True
            self.close_DSOInfo()
            self.update()

    def unpin(self, dt=None):
        if self.pinned:
            if self.ot == 'QS':
                self.text = ' '
            else:
                self.text = '¤'
            self.pinned = False
            self.display_DSOInfo()
            self.update()

    def close_DSOInfo(self, *args):
        dsopopup.hide()

    def display_DSOInfo(self, *args):
        dsopopup.display(info=self.info, title_color=self.marker_color)

    def in_eyepiece(self, xc, yc, r2):
        return (xc - self.sx) ** 2 + (yc - self.sy) ** 2 < r2

    def update(self, mapping=None, xc=None, yc=None, r2=None):
        if xc is not None:
            self.xc = xc
        if yc is not None:
            self.yc = yc
        if mapping is not None:
            self.mapping = mapping
        if r2 is not None:
            self.r2 = r2
        self.sx, self.sy = self.mapping(self.px, self.py)
        self.visible = self.in_eyepiece(self.xc, self.yc, self.r2) and self.display

        #  display if visible and set to display
        if self.visible:
            bx, by = self.mapping(self.bx, self.by)
            if self.pinned:
                self.radius = ((self.sx - bx) ** 2 + (self.sy - by) ** 2) ** 0.5
                ang_r = np.radians(self.ang)
                self.lab_x = float(self.sx + self.radius * np.cos(ang_r))
                self.lab_y = float(self.sy + self.radius * np.sin(ang_r))
            else:
                self.lab_x = self.sx
                self.lab_y = self.sy


class AnnotCluster(Annotation):
    pass


class AnnotQuasar(Annotation):
    # a couple of lines a few pix from QSO
    length = NumericProperty(20)  # indicator line length in pixels
    gap = NumericProperty(10)  # gap from quasar to line in pixels


class AnnotFOV(Annotation):
    # represents FOV line
    length = NumericProperty(0)

    def update(self, mapping=None, **args):
        # compute length and update text based on FOV
        # sx is location
        if mapping is not None:
            self.mapping = mapping
        sx, sy = self.mapping(self.px, self.py)
        bx, by = self.mapping(self.bx, self.by)
        arcsecs = ((sx - bx) ** 2 + (sy - by) ** 2) ** 0.5
        self.length = arcsecs * 60
        self.sx = Window.width // 2
        self.sy = 200
        self.lab_x = self.sx
        self.lab_y = self.sy + 20
        self.visible = self.display

    def on_enter(self, *args):
        pass

    def on_touch_down(self, touch):
        return False


class TitleBar(BoxLayout):
    Name = StringProperty('')
    title_color = ListProperty([1, 0, 0, 1])


class PropVal(BoxLayout):
    param = StringProperty('')
    value = StringProperty('')


class DSOPopup(BoxLayout):
    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        Window.add_widget(self)
        self.hide()

    def display(self, info=None, title_color=None):

        # ideally, exclude some of these in Catalogue
        exclude = {'OT', 'Name', 'RA', 'Dec', 'List', 'Con', 'Obs', 'List', 'Added', 'MaxAlt', 'Transit'}
        # there is a subtle bug here which this is going to catch one day
        try:
            self.clear_widgets()
            tb = TitleBar(Name=info.get('Name', ''))
            if title_color is not None:
                tb.title_color = title_color
            self.add_widget(tb)

            for k, v in info.items():
                if str(v) and str(v) != 'nan' and k not in exclude:
                    if type(v) == float:
                        pv = PropVal(param=k, value='{:.3f}'.format(v))
                    else:
                        pv = PropVal(param=k, value=str(v))
                    self.add_widget(pv)

            # position to popup info; not yet working in terms of limiting it to screen
            pos = Window.mouse_pos
            self.x = min(pos[0] + dp(9), Window.width - self.minimum_width - dp(20))
            self.y = min(pos[1], Window.height - self.minimum_height - dp(20))
        except Exception as e:
            Logger.debug('Annotator: DSOPopup.display ({:})'.format(e))

    def hide(self):
        self.x = 10 * Window.width

# ensures that there is only one ie singleton popup
dsopopup = DSOPopup()

class Annotator(Component):

    mag_limit = NumericProperty(0)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        self.annotations = []

    def on_new_object(self):
        self.clear_annotations()
        self.on_mag_limit()

    def clear_annotations(self):
        # remove previous annotations
        gui = App.get_running_app().gui
        if self.has_annotations():
            for k in self.annotations:
                gui.remove_widget(k)
            self.annotations = []

    def has_annotations(self):
        return len(self.annotations) > 0

    def on_mag_limit(self, *args):
        # modify what is visible
        ml = self.mag_limit

        # if limit is below -5, turn off all annotations
        if ml <= -5:
            for annot in self.annotations:
                annot.display = False
            mesg = 'off'
        # if below 0, just turn on pinned annotations
        elif ml < 0:
            for annot in self.annotations:
                annot.display = annot.pinned
            mesg = 'pin'
        # otherwise apply magnitude limit to unpinned cases
        else:
            for annot in self.annotations:
                annot.display = annot.pinned or annot.mag < ml
            mesg = '{:4.1f}'.format(ml)

        self.app.gui.set_prop('mag_limit', 'text', '   {:}'.format(mesg))

        self.update()

    def get_diam(self, ot, info):
        if 'Diam' in info:
            d = info['Diam']
        elif 'diam' in info:
            d = info['diam']
        elif 'Major' in info:
            d = info['Major']
        else:
            d = .2
        if np.isnan(d):
            d = .2 
        return d


    def annotate(self):
        ''' called by platesolver when finished a platesolve
        '''

        self.clear_annotations()
        solver = Component.get('PlateSolver')
        cats = Component.get('Catalogues')

        # rotate view to north
        Component.get('View').orientation = solver.north

        # find objects in FOV from tile large enough to encompass sensor
        w, h = solver.w, solver.h
        deg_per_pixel = float(solver.FOV_h / h)
        fov = 1.5 * max(solver.FOV_w, solver.FOV_h)
        tile = make_tile(solver.tile_ra0, solver.tile_dec0, fov=fov)

        dsos = cats.get_annotation_dsos(tile=tile)

        target = Component.get('DSO').Name

        # add FOV annotator
        self.annotations = [
            AnnotFOV(
                text="1'",
                info={'Name': ''},
                color=[0.6, 0.6, 0.6, 1],
                px=0,
                py=0,
                bx=1 / (3600 * deg_per_pixel),
                by=0,
            )
        ]

        # object type properties and defaults
        OT_props = cats.get_object_types()
        default_props = {'name': '?', 'col': [0, 0, 0], 'text': '¤', 'mag': 0}

        ''' dsos is a dict of names and properties
        '''
        for nm, info in dsos.items():
            px, py = solver.ra_dec_to_pixels(info['RA'], info['Dec'])
            px, py = float(px[0]), float(py[0])
            #  check it is actually within the image dimensions
            if (px >= 0) & (py >= 0) & (px < w) & (py < h):
                # replace OT acronym in info by full OT name
                ot = info['OT']
                props = OT_props.get(ot, default_props)
                # fill in missing props
                for k, v in default_props.items():
                    if k not in props:
                        props[k] = v
                info['Object type'] = props['name']
                # try to get a diameter value in case we need it
                diam = self.get_diam(ot, info)
                ang = 45
                center = True
                if ot in {'OC', 'GC', 'CG', 'GG', 'G+', 'PG'}:
                    cls = AnnotCluster
                    diam *= 1.3  # make circle slightly larger
                    center = False
                elif ot in {'QS', 'BQ', 'DQ'}:
                    cls = AnnotQuasar
                else:
                    cls = Annotation
                    if ot in {'ME'}:
                        diam = .15   # put labels closer
                        ang = 0
                    elif ot in {'PN'}:
                        diam *= 1.5

                # diam is in arcmin
                # print('Obj: {:} OT {:} diam: {:}'.format(info['Name'], ot, diam))
                pixel_rad = (((diam / (2 * 60.0)) / deg_per_pixel)) / 2**.5

                self.annotations += [cls(
                        ot=ot,
                        text=props['text'],
                        px=px, 
                        py=py,
                        info=info,
                        ang=ang,
                        center=center,
                        mag=info.get('Mag', props['mag']),
                        marker_color=props['col'],
                        bx=px + pixel_rad, 
                        by=py + pixel_rad)
                ]

        # create annotation labels
        mapping = Component.get('View').to_parent
        xc, yc = Metrics.get('origin')
        r2 = Metrics.get('inner_radius') ** 2
        gui = App.get_running_app().gui
        for lab in self.annotations:
            gui.add_widget(lab, index=101)
            lab.update(mapping=mapping, xc=xc, yc=yc, r2=r2)
            #  sort out situations where several have same target name
            if lab.info['Name'] == target:
                lab.pin()
        self.on_mag_limit()

    def update(self):

        # called whenever a redraw is required e.g. View or maglimit changes
        if not self.has_annotations():
            return

        mapping = Component.get('View').to_parent
        xc, yc = Metrics.get('origin')
        r2 = Metrics.get('inner_radius') ** 2
        for annot in self.annotations:
            annot.update(mapping=mapping, xc=xc, yc=yc, r2=r2)




''' Native suport for SX Lodestar camera. This only works reliably on the Mac
    at present.
'''

import time
import array
import os
from concurrent.futures import ThreadPoolExecutor
import numpy as np
from functools import partial

from kivy.logger import Logger
from kivy.app import App
from kivy.properties import BooleanProperty
from kivy.clock import Clock

from jocular.component import Component

#  Lodestar (and other SX cameras) command set
SX_CLEAR_PIXELS = 1
SX_READ_PIXELS_DELAYED = 2
SX_READ_PIXELS = 3
SX_RESET = 6

# convert int formats
def convert_int(x, from_type='uint16', to_type='uint8'):
    return np.array([x], dtype=from_type).view(dtype=to_type)


class Camera(Component):

    camera_connected = BooleanProperty(False)

    def is_connected(self):
        return self.camera_connected

    def on_new_object(self):
        self.reset()

    def reset(self):
        self.connect()
        self.last_capture = None
        self.pause()

    def on_close(self):
        pass

    def pause(self):
        # Called by Capture when user pauses. We cancel any pending reads
        if hasattr(self, 'capture_event'):
            if not self.internal_timing:  
                # don't cancel reading if using internal timing (causes crash on LS)
                self.capture_event.cancel()

    def clean_up(self):
        if hasattr(self, 'lodestar'):
            import usb.util

            usb.util.dispose_resources(self.lodestar)
            Logger.debug('Camera: disposed')

    def handle_failure(self, message='camera problem'):
        Logger.error('Camera: handle_failure {:}'.format(message))
        self.camera_connected = False  # to ensure reconnect if that was the issue
        self.last_capture = None
        try:
            self.clean_up()
        except Exception as e:
            Logger.warn('Camera: problem disposing {:}'.format(e))
            self.warn('disconnected?')
        finally:
            if self.on_failure is not None:
                self.on_failure()

    def get_image(self):
        ''' new in v0.5
        '''
        if hasattr(self, 'last_capture'):
            return self.last_capture
        return None    

    # generic capture routine
    def capture_sub(
        self,
        exposure=None,
        on_capture=None,
        on_failure=None,
        binning=None,
        internal_timing=False,
        return_image=False,
        is_bias=False
    ):

        if is_bias:
            self.capture_bias(on_capture=on_capture, on_failure=on_failure)
            return

        # store these so they can be used without argument passing in futures etc
        self.on_capture = on_capture
        self.on_failure = on_failure
        self.exposure = exposure
        self.binning = binning
        self.internal_timing = internal_timing
        self.return_image = return_image
        self.last_capture = None

        # try to connect to camera (will handle if already connected)
        self.connect()

        if not self.camera_connected:
            self.handle_failure('camera not connected')
            return

        try:
            if internal_timing:
                # do internal_timed exposure
                self.start_internal_exposure()
                if return_image:
                    time.sleep(exposure)
                    return self.lodestar_read()
                else:
                    self.capture_event = Clock.schedule_once(
                        self.read_internal_exposure, max(0.1, exposure)
                    )
            else:
                # externally-timed
                self.start_external_exposure()
                if return_image:
                    time.sleep(exposure)
                    return self.read_external_exposure()
                else:
                    self.capture_event = Clock.schedule_once(
                        self.read_external_exposure, max(0.2, self.exposure)
                    )

        except Exception as e:
            self.handle_failure('capture problem {:}'.format(e))

    def capture_bias(self, on_capture=None, on_failure=None):
        # special case for bias as it needs even and odd for internally-timed exposure
        self.on_capture = on_capture
        self.on_failure = on_failure
        try:
            self.connect()
            expo = 0.001
            self.exposure_command(rows='odd', exposure=expo)
            time.sleep(expo)
            odd_pixels = self.lodestar.read(0x82, self.height * self.width, 10000)
            self.exposure_command(rows='even', exposure=expo)
            time.sleep(expo)
            even_pixels = self.lodestar.read(0x82, self.height * self.width, 10000)
            pool = ThreadPoolExecutor(3)
            future = pool.submit(
                partial(self.deinterlace, odd_pixels, even_pixels)
            )  # thread handles read
            future.add_done_callback(
                self.image_ready
            )  # when future is done, call image_ready
        except Exception as e:
            self.handle_failure('problem in capture_bias {:}'.format(e))

    def exposure_command(self, rows='odd', exposure=None):
        # generate write command for sensor
        try:
            rowcode = {'odd': 1, 'even': 2, 'all': 3}
            cmd = SX_READ_PIXELS if exposure is None else SX_READ_PIXELS_DELAYED
            params = array.array(
                'B',
                [
                    0x40, cmd, rowcode[rows], 0, 0, 0, 10, 0, 0, 0, 0, 0,
                    self.lswidth, self.mswidth, self.lsheight, self.msheight, 1, 1
                ],
            )
            if exposure is not None:
                # add exposure details to command
                expo = int(exposure * 1000)  #  ms
                exp1, exp2, exp3, exp4 = convert_int(expo, from_type='uint32')
                params = params + array.array('B', [exp1, exp2, exp3, exp4])
            self.lodestar.write(1, params, 1000)
        except Exception as e:
            self.handle_failure('problem in exposure_command {:}'.format(e))

    def start_internal_exposure(self):
        try:
            self.exposure_command(rows='odd', exposure=self.exposure)
        except Exception as e:
            self.handle_failure('problem in start_internal_exposure {:}'.format(e))

    def start_external_exposure(self):
        try:
            self.lodestar.ctrl_transfer(0x40, SX_RESET, 0, 0, 0)
            self.lodestar.ctrl_transfer(0x40, SX_CLEAR_PIXELS, 0, 0, 0)
            Logger.trace('Camera: started external exposure')
        except Exception as e:
            self.handle_failure('problem in start_external_exposure {:}'.format(e))

    def read_internal_exposure(self, *args):
        try:
            pool = ThreadPoolExecutor(3)
            future = pool.submit(self.lodestar_read)  # thread handles read
            future.add_done_callback(
                self.image_ready
            )  # when future is done, call image_ready
        except Exception as e:
            self.handle_failure('problem in read_internal_exposure {:}'.format(e))

    def read_external_exposure(self, *args):
        # Reads sensor at end of exposure time

        try:
            self.exposure_command(rows='odd')
            odds = self.lodestar.read(0x82, self.height * self.width, timeout=10000)
            self.exposure_command(rows='even')
            evens = self.lodestar.read(0x82, self.height * self.width, timeout=10000)

            # de-interlace to create final image
            self.last_capture = self.deinterlace(odds, evens)

            if self.on_capture is not None:
                self.on_capture()
        except Exception as e:
            self.handle_failure('problem in read_external_exposure {:}'.format(e))

    def lodestar_read(self):
        try:
            odd_pixels = self.lodestar.read(0x82, self.height * self.width, 10000)
            return self.deinterlace(odd_pixels, odd_pixels)
        except Exception as e:
            self.handle_failure('problem in lodestar_read {:}'.format(e))

    def image_ready(self, future):
        self.last_capture = future.result()  # store image read from sensor
        if self.on_capture is not None:
            Clock.schedule_once(self.on_capture, 0)

    def deinterlace(self, odd8, even8):
        '''De-interlaces Lodestar.'''

        # convert uint8 arrays to uint16 and reshape
        odd = np.frombuffer(odd8, dtype='uint16').reshape(self.half_height, self.width)
        even = np.frombuffer(even8, dtype='uint16').reshape(
            self.half_height, self.width
        )

        # generate new array (full height)
        pix = np.zeros((self.height, self.width))

        # insert odd rows
        pix[::2, :] = odd

        # insert even rows and normalise to account for slight delay in reading
        pix[1::2, :] = even * (np.mean(odd) / np.mean(even))

        return pix / 2 ** 16

    def connect(self):
        """Connect to Lodestar. For other cameras will need to change idProduct
        and check interface/endpoint details.
        """

        if self.camera_connected:
            return

        if os.name == 'nt':
            self.info('not supported')
            return

        # os.environ['PYUSB_DEBUG'] = 'debug'

        # try:
        #     import usb.backend.libusb0
        #     b = usb.backend.libusb0.get_backend()
        # except:
        #     self.warn('no usb.backend.libusb0')
        #     self.camera_connected = False
        #     return

        self.camera_connected = False

        try:
            import usb.core
        except:
            self.warn('no usb.core')
            return

        try:
            import usb.util
        except:
            self.warn('no usb.util')
            return

        self.lodestar = None
        try:
            self.lodestar = usb.core.find(idVendor=0x1278, idProduct=0x0507)
        except:
            self.warn('cannot find cam')

        #  next try libusb1
        if self.lodestar is None and os.name == 'nt':
            Logger.debug('Camera: looking for libusb on Windows')
            # try forcing backend (for windows)
            try:
                libusb_path = App.get_running_app().get_path('libusb')
                backend = usb.backend.libusb1.get_backend(
                    find_library=lambda x: libusb_path
                )
                self.lodestar = usb.core.find(
                    idVendor=0x1278, idProduct=0x0507, backend=backend
                )
            except:
                self.warn('cannot find w/ libusb1')

        if self.lodestar is None:
            self.info('not connected')
            return

        try:
            conf = self.lodestar.get_active_configuration()
        except Exception as e:
            Logger.warn('Camera: cannot get config ({:})'.format(e))
            self.warn('no config')
            self.clean_up()
            return

        # (might restrict this to Windows) try to claim interface
        # causes a crash on windows
        # interface = 0
        # if self.lodestar.is_kernel_driver_active(interface):
        #     Logger.debug('Camera: detaching and claiming interface')
        #     self.lodestar.detach_kernel_driver(interface)
        #     usb.util.claim_interface(self.lodestar, interface)

        # we don't use endpoint but for other cameras we will need code like this to
        #  find the endpoint (offline prior to implementation)

        # ep = camera[0].interfaces()[0].endpoints()[0]

        self.camera_connected = True
        self.info('SX Lodestar')

        # used in capture; note h // 2 for interlaced; needs changing for ultrastar no doubt
        self.width, self.height = 752, 580
        self.half_height = self.height // 2
        self.lswidth, self.mswidth = convert_int(self.width)
        self.lsheight, self.msheight = convert_int(self.half_height)




class SXCamera(Camera):
    pass

class WatchedCamera(Camera):
    pass






''' Represents session info
'''

import json
from datetime import datetime

from kivy.logger import Logger
from kivy.app import App
from kivy.properties import StringProperty, BooleanProperty, ConfigParserProperty, NumericProperty
from kivy.clock import Clock
from kivy.lang import Builder
from kivy.uix.boxlayout import BoxLayout
from kivy.metrics import dp

from jocular.component import Component
from jocular.widgets import JPopup

Builder.load_string('''

<MyLabel@Label>:
    size_hint: 1, None
    height: dp(24)
    text_size: self.size
    markup: True
    halign: 'left'
    font_size: app.form_font_size

<MyTextInput@TextInput>:
    size_hint: 1, None
    background_color: app.background_color
    foreground_color: app.lowlight_color
    multiline: True
    hint_text_color: app.lowlight_color
    height: dp(28)
    cursor_width: '2sp'
    font_size: app.form_font_size

<ParamValue1@ParamValue>:
    callback: root.parent.edit

<Session>:
    size_hint: None, None
    orientation: 'vertical'
    padding: dp(5), dp(2)
    size: max(dp(250), (app.gui.width - app.gui.height) / 2), app.gui.height / 2
    y: 0
    x: 0 if root.show_session else -self.width

    Button:
        size_hint: 1, 1
        markup: True
        text_size: self.size
        padding: dp(3), dp(1)
        halign: 'left'
        valign: 'bottom'
        font_size: app.info_font_size
        background_color: 0, 0, 0, 0
        color: app.lowlight_color
        height: dp(20)
        text: '[b]Session notes[/b] {:}'.format(root.session_notes)

    ParamValue1:
        param: 'Date'
        value: root.session
    ParamValue1:
        param: 'Temp.'
        value: '' if root.temperature is None else '{:}\N{DEGREE SIGN}C'.format(root.temperature)
    ParamValue1:
        param: 'SQM'
        value: '' if root.SQM is None else '{:.2f} mag/arcsec^2'.format(root.SQM)
    ParamValue1:
        param: 'Seeing'
        value: '{:}'.format(root.seeing if root.seeing else '')
    ParamValue1:
        param: 'Transp.'
        value: '{:}'.format(root.transparency if root.transparency else '')
    ParamValue1:
        param: 'Scope'
        value: root.telescope
    ParamValue1:
        param: 'Camera'
        value: root.camera

<SessionInfo>:
    orientation: 'vertical'
    size_hint: None, None
    size: dp(300), dp(500)
    spacing: dp(5)

    MyLabel:
        text: 'Session notes:'
    MyTextInput:
        id: _notes
        hint_text: 'not specified'
        text: root.session.session_notes
        multiline: True
        size_hint: 1, None
        height: dp(80)
        on_text: root.session.session_notes = self.text

    MyLabel:
        text: 'Seeing:'
    MyTextInput:
        id: _seeing
        hint_text: ' not specified'
        text: root.session.seeing
        on_text: root.session.seeing = self.text

    MyLabel:
        text: 'Transparency:'
    MyTextInput:
        id: _transparency
        hint_text: 'not specified'
        text: root.session.transparency
        on_text: root.session.transparency = self.text

    MyLabel:
        text: 'Temperature {:}\N{DEGREE SIGN}C'.format(_temperature.value) if _temperature.value > -26 else 'Temperature not set'
    Slider:
        size_hint: 1, None
        height: dp(30)
        id: _temperature
        value: root.session.temperature if root.session.temperature is not None else -26
        on_value: root.session.temperature = self.value if self.value > -26 else None
        min: -26
        max: 40
        step: 1

    MyLabel:
        text: 'SQM {:.2f} mag/arcsec^2'.format(_sqm.value) if _sqm.value > 15.9999 else 'SQM not set'
    Slider:
        size_hint: 1, None
        height: dp(24)
        id: _sqm
        value: root.session.SQM if root.session.SQM is not None else 15.9999
        on_value: root.session.SQM = self.value if self.value > 16 else None
        min: 15.9
        max: 22
        step: .01

    MyLabel:
        text: 'Telescope:'
    MyTextInput:
        id: _scope
        hint_text: 'not specified'
        text: root.session.telescope
        on_text: root.session.telescope = self.text

    MyLabel:
        text: 'Camera:'
    MyTextInput:
        id: _camera
        hint_text: 'not specified'
        text: root.session.camera
        on_text: root.session.camera = self.text

    BoxLayout:
        size_hint: 1, None
        height: dp(40)
        Button:
            text: 'Close'
            size_hint: 1, .8
            on_press: root.session.done()

''')

class SessionInfo(BoxLayout):
    def __init__(self, session, **kwargs):
        self.session = session
        super().__init__(**kwargs)

date_time_format = '%d %b %y %H:%M'

def datenow():
    return datetime.now().strftime(date_time_format)

def hours_since_date(mydate):
    return (datetime.now() - datetime.strptime(mydate, date_time_format)).total_seconds()/3600

class Session(BoxLayout, Component):

    session = StringProperty('')
    SQM = NumericProperty(None, allownone=True)
    temperature = NumericProperty(None, allownone=True)
    seeing = StringProperty('')
    transparency = StringProperty('')
    session_notes = StringProperty('')
    telescope = StringProperty('')
    camera = StringProperty('')

    retain_hours =  ConfigParserProperty(3, 'Session', 'retain_hours', 'app', val_type=float)
    is_new_object = BooleanProperty(False)
    show_session = BooleanProperty(False)

    props = ['SQM', 'temperature', 'seeing', 'transparency', 'session_notes', 
        'telescope', 'camera', 'session']

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        self.app.gui.add_widget(self, index=2) 

    def on_new_object(self):
        self.is_new_object = not Component.get('ObjectIO').existing_object
        # move this to see if I can cancel session clock more reliably
        if hasattr(self, 'session_clock'):
            self.session_clock.cancel()
        if self.is_new_object:
            self.load_session()
        else:
            settings = Component.get('Metadata').get(self.props)
            self.apply_settings(settings)

    def load_session(self):
        try:
            with open(self.app.get_path('last_session.json'), 'r') as f:
                session = json.load(f)
            if hours_since_date(session['session']) < self.retain_hours:
                self.apply_settings(session)
            else:
                # start fresh session but restore scope/camera details
                self.create_new_session()
                self.telescope = session.get('telescope', '') 
                self.camera = session.get('camera', '')
        except Exception as e:
            # any problems we just start a fresh session
            Logger.debug('Session: problem loading session data, so recreating {:}'.format(e))
            self.create_new_session()

        # create date and clock to update it
        self.session = datetime.now().strftime(date_time_format)
        self.session_clock = Clock.schedule_interval(self.update_time, 1.0)

    def create_new_session(self):
        self.SQM = None
        self.temperature = None
        self.seeing = ''
        self.transparency = ''
        self.session_notes = ''
        self.telescope = ''
        self.camera = ''

    def get_session(self):
        # accessor method that strips off seconds if present
        if self.session.count(':') == 2:
            return self.session[:-3]
        else:
            return self.session

    def current_session_info(self):
        session = {}
        for p in self.props:
            v = getattr(self, p)
            if v:
                session[p] = v
        session['session'] = self.get_session()
        return session

    def apply_settings(self, settings):
        self.create_new_session()
        for p in self.props:
            if p in settings:
                setattr(self, p, settings[p])

    def save_session(self):
        Logger.debug('Session: saving session')
        with open(self.app.get_path('last_session.json'), 'w') as f:
            json.dump(self.current_session_info(), f, indent=1)    

    def on_save_object(self):

        # save properties
        for p in ['SQM', 'temperature', 'seeing', 'transparency', 'session_notes', 'telescope', 'camera']:
            Component.get('Metadata').set(p, getattr(self, p))

        # only update last session if live rather than previous
        if self.is_new_object:
            self.save_session()

    def update_time(self, *args):
        self.session = datetime.now().strftime('%d %b %y %H:%M:%S')

    def edit(self, *args):
        content = SessionInfo(self)
        self.popup = JPopup(title='Session information', content=content, posn='bottom-left')
        self.popup.open()

    def done(self, *args):
        self.popup.dismiss()
        self.changed = not Component.get('Stacker').is_empty() and (
            self.temperature != Component.get('Metadata').get('temperature', default=None) or \
            self.SQM != Component.get('Metadata').get('SQM', default=None) or \
            self.seeing.strip() != Component.get('Metadata').get('seeing', default='') or \
            self.transparency.strip() != Component.get('Metadata').get('transparency', default='') or \
            self.telescope.strip() != Component.get('Metadata').get('telescope', default='') or \
            self.camera.strip() != Component.get('Metadata').get('camera', default='')
            )
        if self.is_new_object:
            self.save_session()

    def on_touch_down(self, touch):
        # print(' TD in Session')
        if self.collide_point(*touch.pos) and touch.pos[0] < dp(100) and self.app.showing == 'main':
            return super().on_touch_down(touch)
        return False



''' Manages status panel showing information about a subset of components
'''

from kivy.app import App
from kivy.lang import Builder
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.label import Label
from kivy.properties import BooleanProperty

from jocular.component import Component

Builder.load_string('''

<StatusLabel>:
    size_hint: 1, None
    height: dp(20) 
    halign: 'right'
    markup: True
    font_size: app.info_font_size
    text_size: self.size
    size: self.texture_size
    padding: dp(3), dp(1)
    color: app.lowlight_color
    shorten: True

<Status>:
    size_hint: None, None
    orientation: 'vertical'
    padding: dp(5), dp(2)
    size: app.gui.height / 8 + (app.gui.width - app.gui.height) / 2, app.gui.height / 2
    pos: app.gui.width - self.width if root.show_status else app.gui.width, 0
    Label:
        size_hint: 1, 1
''')

class StatusLabel(Label):
    pass

class Status(BoxLayout, Component):

    show_status = BooleanProperty(False)
    comps = ['ObjectIO', 'BadPixelMap', 
             'Calibrator', 'View', 'Observations',
             'ObservingList', 'Aligner', 'Stacker', 'PlateSolver', 'Snapshotter', 'FilterWheel', 'Capture']

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        App.get_running_app().gui.add_widget(self, index=2)

        # add a status label for each component where we are interested in status updates
        self.labs = {c: StatusLabel(text=c) for c in self.comps}
        for name, lab in self.labs.items():
            self.add_widget(lab)

    def bind_status(self, name, attr):
        # called by Component when a component is loaded
        if name in self.labs:
            attr.bind(infoline=self.labs[name].setter('text'))



''' Observing notes panel.
'''

from kivy.clock import Clock
from kivy.app import App
from kivy.properties import StringProperty, BooleanProperty, ObjectProperty
from kivy.lang import Builder
from kivy.uix.boxlayout import BoxLayout

from jocular.component import Component
from jocular.widgets import JPopup

Builder.load_string('''
<Notes>:
    canvas.before:
        Color:
            rgba: 1, 0, 0, 0
        Rectangle:
            pos: self.pos
            size: self.size
    size_hint: None, None
    orientation: 'vertical'
    padding: dp(5), dp(2)
    size: ((app.gui.width - app.gui.height) / 2) - dp(4), app.gui.height / 3
    y: 2 * app.gui.height / 3
    x: app.gui.width - self.width if root.show_notes else app.gui.width

    Button:
        text: 'Observational notes: {:}'.format(root.notes)
        # size: self.texture_size
        background_color: .6, 0, .6, 0
        size_hint: 1, 1
        valign: 'top'
        halign: 'right'
        color: app.lowlight_color
        on_press: root.edit()
        markup: True
        font_size: app.info_font_size
        text_size: self.size

<NotesInfo>:
    orientation: 'vertical'
    size_hint: None, None
    size: dp(300), dp(360)
    spacing: dp(5)
    notes_input: _notes

    TextInput:
        id: _notes
        unfocus_on_touch: False
        background_color: app.background_color
        foreground_color: app.lowlight_color
        hint_text_color: app.lowlight_color
        hint_text: 'type your observing notes here'
        text: root.notes.notes
        multiline: True
        size_hint: 1, None
        height: dp(300)
        font_size: app.form_font_size
        valign: 'top'

    BoxLayout:
        size_hint: 1, None
        height: dp(40)
        Button:
            text: 'Save'
            size_hint: .5, .8
            on_press: root.notes.edited(_notes.text)
        Button:
            text: 'Cancel'
            size_hint: .5, .8
            on_press: root.notes.cancel_edit()

''')

class NotesInfo(BoxLayout): 
    notes_input = ObjectProperty(None)
    def __init__(self, notes, **kwargs):
        self.notes = notes
        super().__init__(**kwargs)
        Clock.schedule_once(self.set_focus, 0)

    def set_focus(self, dt):
        self.notes_input.focus = True

class Notes(BoxLayout, Component):

    notes = StringProperty('')
    show_notes = BooleanProperty(False)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        App.get_running_app().gui.add_widget(self) # , index=2) 

    def on_new_object(self):
        self.notes =  Component.get('Metadata').get('Notes', default='')

    def on_save_object(self):
        if len(self.notes) > 0:
            Component.get('Metadata').set('Notes', self.notes)

    def edit(self, *args):
        content = NotesInfo(self)
        self.popup = JPopup(title='Observing notes', content=content, posn='top-right')
        self.popup.open()
        
    def edited(self, notes):
        self.notes = notes
        self.changed = self.notes != Component.get('Metadata').get('Notes', default='') and \
            not Component.get('Stacker').is_empty()
        self.popup.dismiss()

    def cancel_edit(self, *args):
        self.popup.dismiss()



''' Manages previous observations and associated observations table.
'''

import os
import random
import shutil
import time
import glob
import json
from datetime import datetime

from kivy.app import App
from kivy.clock import Clock
from kivy.core.window import Window
from kivy.logger import Logger

from jocular.component import Component
from jocular.table import Table
from jocular.metadata import get_metadata


class Observations(Component):

    def __init__(self):
        super().__init__()
        self.app = App.get_running_app()
        Clock.schedule_once(self.load_observations, 0)
        # self.load_observations()

    def get_observations(self):
        if not hasattr(self, 'observations'):
            self.load_observations()
        return self.observations

    def load_observations(self, dt=None):
        ''' Load json file containing observation details. If any problem,
            set observations to empty and continue.
        '''
        try:
            with open(self.app.get_path('previous_observations.json'), 'r') as f:
                self.observations = json.load(f)
            Logger.info('Observations: Loaded {:} observations'.format(
                len(self.observations)))
        except Exception as e:
            Logger.warn('Observations: none loaded ({:})'.format(e))
            self.observations = {}
        self.info('Loaded {:}'.format(len(self.observations)))

    def save_observations(self):
        try:
            with open(self.app.get_path('previous_observations.json'), 'w') as f:
                json.dump(self.observations, f, indent=1)
            Logger.info('Observations: saved {:}'.format(len(self.observations)))
        except Exception as e:
            Logger.warn('Observations: cannot save previous observations ({:})'.format(e))

    def build_observations(self):
        # table construction ~150ms

        return Table(
            size=Window.size,
            data=self.get_observations(),
            name='Observations',
            description='click on DSO name to load',
            cols={
                'Name': {'w': 250, 'align': 'left', 'sort': {'catalog':''}, 'action': self.load_dso},
                'OT': {'w': 40},
                'Con': {'w': 50},
                'Session': {'w': 140, 'sort': {'DateFormat': '%d %b %y %H:%M'}},
                'N': {'w': 40, 'type': int},
                'Notes': {'w': 1, 'align': 'left'}
                },
            actions={'move to delete dir': self.move_to_delete_folder},
            on_hide_method=self.app.table_hiding,
            initial_sort_column='Session', 
            initial_sort_direction='reverse'
            )    

    def show_observations(self, *args):
        '''Called from menu to browse DSOs; open on first use
        '''

        if not hasattr(self, 'observations_table'):
            self.observations_table = self.build_observations()
        self.app.showing = 'observations'

        # redraw on demand as it is expensive
        if self.observations_table not in self.app.gui.children:
            self.app.gui.add_widget(self.observations_table, index=0)

        self.observations_table.show()    

    def on_close(self, *args):
        ''' Save observations, clean up any empty observation and session dirs
        '''

        self.save_observations()
        sesh = Component.get('ObjectIO').session_dir
        for dso in glob.glob(os.path.join(sesh, '*')):
            if os.path.isdir(dso):
                # remove any empty observation directories
                if len(os.listdir(dso)) == 0:
                    try:
                        os.rmdir(dso)
                    except Exception as e:
                        Logger.warn('Observations: cannot remove observation directory {:} ({:})'.format(dso, e))
        # remove empty session dirs
        for sesh in glob.glob(os.path.join(self.app.get_path('captures'), '*')):
            if len(os.listdir(sesh)) == 0:
                try:
                    os.rmdir(sesh)
                except Exception as e:
                    Logger.warn('Observations: cannot remove session directory {:} ({:})'.format(sesh, e))

    def move_to_delete_folder(self, *args):
        for s in self.observations_table.selected:
            if s in self.observations:
                self._delete(s)
                del self.observations[s]
        self.observations_table.update()
        self.save_observations()
        self.info('{:}'.format(len(self.observations)))

    def _delete(self, path):
        try:
            dname = os.path.join(self.app.get_path('deleted'),
                '{:}_{:}_{:d}'.format(os.path.basename(path),
                datetime.now().strftime('%d_%b_%y_%H_%M'),
                random.randint(1,9999)))
            shutil.move(os.path.join(self.app.get_path('captures'), path), dname)
        except Exception as e:
            Logger.error('Observations: deleting observations ({:})'.format(e))
            self.error('problem on delete')

    def update(self, oldpath, newpath):
        ''' We have either an existing observation being saved or a 
            new observation, so read it and generate extra required info to 
            update self.observations
        '''

        # if not yet built, no worries because it will be found when next built
        if not hasattr(self, 'observations'):
            return

        self.observations[newpath] = get_metadata(newpath, simple=True)
        # user has changed directory
        if oldpath != newpath:
            if oldpath in self.observations:
                del self.observations[oldpath]

        # only update if we have already displayed table
        if hasattr(self, 'observations_table'):
            self.observations_table.update()

        self.save_observations()
        self.info('{:}'.format(len(self.observations)))

    def load_dso(self, row):
        self.observations_table.hide()
        # convert row.key to str (it is numpy.str by default)
        Component.get('ObjectIO').load_previous(path=str(row.key))



''' Runs the capture scripts using generators.

    New in v0.5 is the selection of different capture modes e.g. 
    watched folder, native SX, ... with more to come. These are
    handled by the on_capture_mode method, otherwise no real
    changed here. Most of the changes are in Camera, which
    implements a subclass for each mode.
'''

import os
import time
import numpy as np
from scipy.interpolate import interp1d

from kivy.app import App
from kivy.properties import NumericProperty, BooleanProperty, ConfigParserProperty
from kivy.clock import Clock
from kivy.logger import Logger

from jocular.component import Component
from jocular.widgets import JPopup
from jocular.image import save_image
from jocular.gradient import image_stats
from jocular.camera import (
    GenericCamera, SXCamera, WatchedCamera, WatchedCameraOnOff,
    AscomCamera, ASICamera
    )

class Capture(Component):

    capturing = BooleanProperty(False)
    exposure = NumericProperty(0)
    scripts = [s + '_script' for s in {'seq', 'light', 'frame', 'dark', 'flat', 'bias'}]
    capture_mode = ConfigParserProperty(
        'watcher', 'Capture Modes', 'capture_mode', 'app', val_type=str)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        self.gui = self.app.gui
        self.camera = GenericCamera(self)
        self.last_mode = None
        self.modes = {}

    def on_new_object(self):
        self.reset()
        self.on_capture_mode() # set up capture mode
        self.camera.on_new_object()

    def on_capture_mode(self, *args):
        ''' Factory-style instantiation of object to represent camera.
            Caches objects for fast switching.
        '''

        #for k, v in self.modes.items():
        #    print('  mode {:} connected {:} active {:}'.format(k, v.is_connected(), k==self.capture_mode)) 

        if self.last_mode is not None and self.capture_mode == self.last_mode:
            return

        mode = self.capture_mode
        if self.last_mode is not None:
            self.modes[self.last_mode].deselect()
        self.last_mode = mode

        # if we already instantiated this mode, switch camera to it
        if mode in self.modes:
            self.camera = self.modes[mode]
            self.camera.select()
            Logger.info('Capture: Capture mode changed to cached {:}'.format(mode))
            self.info(mode)
            return

        # instantiate this mode if it is one we can handle
        if mode == 'native SX':
            c = SXCamera(self)
        elif mode == 'ASCOM':
            c = AscomCamera(self)
        elif mode == 'ASI':
            c = ASICamera(self)
        elif mode == 'watcher':
            c = WatchedCamera(self)
        elif mode == 'controllable watcher':
            c = WatchedCameraOnOff(self)
        else:
            Logger.info('Capture: Mode {:} not implemented'.format(mode))
            return

        Logger.info('Capture: Capture mode changed to {:}'.format(mode))
        self.info(mode)
        self.modes[mode] = c
        self.camera = c
        self.camera.select()

    def connect(self):
        self.camera.connect()

    def is_connected(self):
        return self.camera.is_connected()

    def pause(self):
        self.camera.pause()

    def reset(self):
        # to do: the camera should reset some of these things depending on its mode
        # to do: check this as it incorporates stuff from capture as well as camera 
        self.series_number = None
        self.stop_capture()
        self.gui.set('capturing', False)
        self.gui.enable(self.scripts + ['exposure_button', 'filter_button'])
        self.camera.reset()

    def get_image(self):
        return self.camera.get_image()

    def capture_sub(self, exposure=None, on_capture=None, on_failure=None, binning=None, 
                    internal_timing=False, return_image=False, is_bias=False):
        return self.camera.capture_sub(exposure=exposure, on_capture=on_capture,
                on_failure=on_failure, binning=binning, internal_timing=internal_timing,
                return_image=return_image, is_bias=is_bias)

    def on_close(self):
        self.camera.on_close()

    def disable_captures(self):
        self.gui.disable(self.scripts + ['capturing', 'exposure_button', 'filter_button'])

    def enable_captures(self):
        self.gui.enable(self.scripts + ['capturing', 'exposure_button', 'filter_button'])

    def soft_reset(self):
        # when stack is cleared we dont want full reset
        self.series_number = None

    def on_capturing(self, *args):
        # user (pressing camera button) or system changes capture state

        self.info('capturing' if self.capturing else 'paused')
        current_script = Component.get('CaptureScript').current_script

        if self.capturing:
            # user wants to capture, so check that camera is connected
            Logger.debug('Capture: check if camera is connected')
            if not self.camera.is_connected():
                self.camera.connect()
            if not self.camera.is_connected():
                self.gui.set('capturing', False, update_property=True)
                Logger.debug('Capture: cannot connect')
                # JBubble(message='No camera connected').open()
                return

            Logger.debug('Capture: camera connected, starting capture')

            # force user to pause/stop before changing any capture parameters
            self.gui.disable(self.scripts + ['exposure_button', 'filter_button', 'load_previous', 'new_DSO'])
            # enable current script so user can see what is being captures
            self.gui.enable(['{:}_script'.format(current_script)])
            try:
                self.capture()
            except Exception as e:
                Logger.error('Capture: problem capturing ({:})'.format(e))

        else:
            self.camera.pause()

            # we can always do framing
            self.gui.enable(['frame_script', 'exposure_button', 'filter_button', 'new_DSO'])
            # find out what stack contains
            subs = Component.get('Stacker').subs
            if len(subs) > 0:
                current = subs[0].sub_type
                if current == 'light':
                    self.gui.enable(['light_script', 'seq_script'])
                else:
                    self.gui.enable(['{:}_script'.format(current)])
            else:
                self.gui.enable(self.scripts + ['load_previous'])                   


    def stop_capture(self, *args):
        # called when problem capturing
        self.gui.set('capturing', False, update_property=True)
        self.gui.enable(self.scripts + ['exposure_button', 'filter_button'])

    def slew(self, *args):
        Logger.info('Capture: slew not implemented yet')

    def camera_disconnected(self):
        self.stop_capture()
        # temporarily comment this out as it should be done by the camera
        # self.gui.disable(['capturing'])

    def camera_reconnected(self):
        # temp comment out
        # self.gui.enable(['capturing'])
        pass

    def capture(self, *args):
        # generator yields next command to execute

        # check if user has pressed pause/stop
        if not self.capturing:
            return 

        # get next command from generator
        op = next(Component.get('CaptureScript').generator)

        if len(op) == 2:
            op, param = op

        # automatically change filter wheel, or request user to do so
        if op == 'set filter':
            Component.get('FilterWheel').select_filter(name=param, 
                changed_action=self.capture,
                not_changed_action=self.stop_capture)

        elif op == 'set exposure':
            self.exposure = param
            self.capture()

        # carry out a normal exposure
        elif op == 'expose long':
            try:
                self.info('capturing ...')
                self.camera.capture_sub(
                    exposure=self.exposure, 
                    on_capture=self.save_capture,
                    on_failure=self.stop_capture)
                if self.exposure > 2:
                    self.expo = self.exposure
                    self.exposure_start_time = time.time()
                    Clock.schedule_once(self.tick, 0)
                self.info('capturing')
            except Exception as e:
                Logger.error('Capture: problem in expose long {:}'.format(e))

        elif op == 'expose short':
            # we update info twice to ensure display updates...
            try:
                self.info('shorts ...')
                self.camera.capture_sub(
                    exposure=self.exposure, 
                    on_capture=self.send_to_display,
                    on_failure=self.stop_capture,
                    internal_timing=self.exposure < 1)
                self.info('shorts')
            except Exception as e:
                Logger.error('Capture: problem in expose short {:}'.format(e))

        elif op == 'expose bias':
            self.camera.capture_sub(
                on_capture=self.save_capture,
                on_failure=self.stop_capture,
                is_bias=True)
            self.info('capturing bias')

        elif op == 'autoflat':
            auto_expo = self.get_flat_exposure()
            if auto_expo is not None:
                self.exposure = auto_expo
                self.capture()
            else:
                self.stop_capture()
                # self.capturing = False

    def on_exposure(self, *args):
        Component.get('CaptureScript').set_exposure_button(self.exposure)        

    # move this to camera or count up?
    def tick(self, *args):
        # remaining = self.expo - (time.time() - self.exposure_start_time)
        dur = time.time() - self.exposure_start_time
        if self.capturing:
            self.info('Exposing {:2.0f}s'.format(dur))
            Clock.schedule_once(self.tick, 1)

    def send_to_display(self, *args):
        # send short subs directly to display
        try:
            Component.get('Monochrome').display_sub(self.camera.get_image())
            self.capture()
        except Exception as e:
            Logger.error('Capture: problem send to display {:}'.format(e))

    def get_capture_details(self):
        sub_type = Component.get('CaptureScript').current_script
        if sub_type == 'seq':
            sub_type = 'light'
        return {
            'exposure': self.exposure,
            'filter': Component.get('FilterWheel').current_filter,
            'sub_type': sub_type,
            'temperature': Component.get('Session').temperature
            }

    def save_capture(self, *args):
        ''' Called via camera when image is ready.
        '''

        # im = Component.get('Camera').get_image()
        im = self.camera.get_image()

        if im is None:
            self.warn('No image')
            return

        # we'll save as 16-bit int FITs
        im *= (2**16 - 1)

        details = self.get_capture_details()
        if not hasattr(self, 'series_number') or self.series_number is None:
            self.series_number = 1
        else: 
            self.series_number += 1

        sub_type = details['sub_type']

        filt = Component.get('FilterWheel').current_filter
        pref = sub_type if sub_type in {'flat', 'dark'} else details['filter']
        name = '{:}_{:d}.fit'.format(pref, self.series_number)

        save_image(data=im.astype(np.uint16),
            path=os.path.join(self.app.get_path('watched'), name),
            exposure=details['exposure'],
            filt=filt,
            temperature=details['temperature'],
            sub_type=sub_type)

        # ask for next capture immediately
        self.capture()

    def get_flat_exposure(self):
        # make a series of test exposures to get ADU just over half-way (0.5)

        min_exposure, max_exposure = 1, 2.5
        min_ADU, max_ADU = .3, .8

        expos = np.linspace(min_exposure, max_exposure, 5)
        adus = np.ones(len(expos))      # normalised ADUs actually
        for i, expo in enumerate(expos):
            # im = Component.get('Camera').capture_sub(exposure=expo, 
            im = self.camera.capture_sub(exposure=expo, 
                return_image=True,
                internal_timing=True,
                on_failure=self.stop_capture)

            # analyse image
            stats = image_stats(im)
            adus[i] = stats['central 75%']
            Logger.debug('Capture: autoflat exposure {:.1f}s has ADU of {:.2f}'.format(expo, adus[i]))

        # are any within ADU tolerance?
        if np.min(adus) > max_ADU:
            JPopup(title='Too early to collect flats', cancel_label='close').open()
            return None

        if np.max(adus) < min_ADU:
            JPopup(title='Too late to collect flats', cancel_label='close').open()
            return None

        # we are OK, so interpolate to get things purrrrfect
        f = interp1d(expos, adus)
        adu_target = .7
        xvals = np.linspace(min_exposure, max_exposure, 500)
        best = np.argmin(np.abs(adu_target - f(xvals)))
        best_exposure = xvals[best]
        Logger.debug('Capture: best exposure for autoflat {:} with ADU {:}'.format(best_exposure, f(best_exposure)))
        return float(best_exposure)



''' Saves currently displayed image in various formats.
'''

import mss
import os
from datetime import datetime
import numpy as np
from skimage.io import imsave
from PIL import Image, ImageDraw, ImageFont

from kivy.logger import Logger
from kivy.app import App
from kivy.core.window import Window
from kivy.properties import OptionProperty, ConfigParserProperty
from kivy.metrics import Metrics as KivyMetrics

from jocular.component import Component
from jocular.image import save_image
from jocular.metrics import Metrics
from jocular.utils import s_to_minsec, unique_member, purify_name

class Snapshotter(Component):

    style = OptionProperty('eyepiece', options=['eyepiece', 'landscape'])
    annotation = OptionProperty('full', options=['plain', 'name', 'full'])
    save_format = OptionProperty('.png', options=['png', 'jpg', 'fit'])
    dark_surround = ConfigParserProperty(1, 'Snapshotter', 'dark_surround', 'app', val_type=int)
    framewidth = ConfigParserProperty(1, 'Snapshotter', 'framewidth', 'app', val_type=int)
    imreduction = ConfigParserProperty(1, 'Snapshotter', 'imreduction', 'app', val_type=float)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()

    def on_new_object(self):
        self.info('ready')

    def mss_grabber(self, bbox=None):
        with mss.mss() as sct:
            im = sct.grab(bbox)
            return Image.frombytes('RGB', im.size, im.bgra, 'raw', 'BGRX').convert('RGBA')

    def large_font(self, w):
        f = max(18, int(w / 32))
        font = ImageFont.truetype(self.app.get_path('Roboto-Medium.ttf'), f)
        (width, height), (offset_x, offset_y) = font.font.getsize('A')
        return font, f, height

    def small_font(self, w):
        f = max(14, int(w / 52))
        font = ImageFont.truetype(self.app.get_path('Roboto-Medium.ttf'), f)
        (width, height), (offset_x, offset_y) = font.font.getsize('A')
        return font, f, height

    def grey(self):
        if self.dark_surround:
            return (150, 150, 150, 255)
        else:
            return (50, 50, 50, 255)

    def snap(self, return_image=False, *args):

        self.return_image = return_image

        view = Component.get('View')
        if not hasattr(view, 'last_image') or (view.last_image is None):
            return

        dso = Component.get('DSO')

        self.save_path = '{:} {:}.{:}'.format(
                os.path.join(self.app.get_path('snapshots'), purify_name(dso.Name)), 
                datetime.now().strftime('%d%b%y_%H_%M_%S'),
                self.save_format)

        if not return_image and self.save_format == 'fit':
            self.save_as_fits()

        elif self.style == 'eyepiece':
            im = self.eyepiece_view()

        elif self.style == 'landscape':
            im = self.landscape_view()

        if return_image:
            return im

    def save_as_fits(self):
        stacker = Component.get('Stacker')
        im = stacker.get_stack()
        if im is not None:
            im = im * (2**16 - 1)
            Logger.debug('Snapshotter: min {:} max {:}'.format(np.min(im), np.max(im)))
            save_image(data=im.astype(np.uint16), path=self.save_path,
                exposure=stacker.describe().get('total_exposure', None),
                sub_type=unique_member([s.sub_type for s in stacker.subs]),
                filt = unique_member([s.filter for s in stacker.subs]))
            self.info(os.path.basename(self.save_path)[:-4])
        else:
            self.warn('no image')

    # construct annotations

    def DSO_details(self):
        dso = Component.get('DSO')
        obj_details = []
        if len(dso.OT) > 0 and (dso.OT in dso.otypes):
            if len(dso.Con) > 0:
                obj_details  += ['{:} in {:}'.format(dso.otypes[dso.OT], dso.Con)]
            else:
                obj_details  += [format(dso.otypes[dso.OT])]

        elif len(dso.Con) > 0:
            obj_details += [dso.Con]

        if dso.RA:
            obj_details += ['RA {:}'.format(str(dso.RA))]
        if dso.Dec:
            obj_details += ['Dec {:}'.format(str(dso.Dec))]
        if dso.Mag:
            obj_details += ['mag {:}'.format(dso.Mag)]
        if dso.Diam:
            obj_details += ["diam {:}'".format(dso.Diam)]

        return obj_details

    def kit_details(self):
        s = Component.get('Session')
        return [s.telescope, s.camera]

    def session_details(self):
        s = Component.get('Session')
        return [s.session, 
            '{:5.2f} SQM'.format(s.SQM) if s.SQM else None,
            '{:}\N{DEGREE SIGN}C'.format(s.temperature) if s.temperature else None,
            'seeing {:}'.format(s.seeing) if s.seeing else None, 
            'transp {:}'.format(s.transparency) if s.transparency else None
            ]

    def processing_details(self):

        block = []
        stacker = Component.get('Stacker')
        d = stacker.describe()
        if d is None:
            return []

        if 'total_exposure' in d:
            block += [s_to_minsec(d['total_exposure'])]
            if d['multiple_exposures']:
                block += ['varied exposures']
            else:
                block += ['{:} x {:.0f}s'.format(d['nsubs'], d['sub_exposure'])]

        if not d['filters'].endswith('L'): # more interesting than just LUM
            block += [d['filters']]

        if d['nsubs'] > 1:
            block += ['{:} stack'.format(stacker.combine)]

        block += ['{:} stretch'.format(Component.get('Monochrome').stretch)]

        calib = ','.join([c for c in ['dark', 'flat', 'bias'] if d[c]])
        calib = calib if calib else 'no darks/flats'
        block += [calib]

        return block # block[::-1]  # reverse!

    def eyepiece_view(self):

        ox, oy = Metrics.get('origin')
        radius = Metrics.get('radius')

        # thanks to callump for the use of scrn_density
        scrn_density = KivyMetrics.density
        cx = int(Window.left + ox / scrn_density)
        cy = int(Window.top + Window.height / scrn_density - oy / scrn_density)
        r = int(radius / scrn_density - 40)
        w = 2 * r

        try:
            im = self.mss_grabber(bbox=(cx - r, cy - r, cx + r, cy + r))
        except Exception as e:
            Logger.error('Snapshotter: problem with screen grab ({:})'.format(e))
            self.error('screengrab issue')
            return

        blur = 17                     # radius of blurred edge in pixels

        # new from callump
        w = im.width
        r = w / 2

        # construct alpha channel to composite eyepiece circle
        # quite likely some of these steps are redundant
        alpha = Image.new('L', (w, w), color=0)
        _alpha = np.ones((w, w))
        grid_x, grid_y = np.meshgrid(np.arange(w), np.arange(w))
        d = pow((grid_x - r) ** 2 + (grid_y - r) ** 2, .5) - (r - 30)
        db = d <= blur
        _alpha[db] = d[db] / blur
        alpha.putdata(_alpha.ravel(), scale=255)
        if self.dark_surround:
            grad_im = Image.new('RGBA', (w, w), color=(0, 0, 0))
        else:
            grad_im = Image.new('RGBA', (w, w), color=(255, 255, 255)) 
        grad_im.putalpha(alpha)

        # composite with scope image, converting if needed (to test)
        try:
            if im.mode != 'RGBA':
                im = im.convert('RGBA')
            im = Image.alpha_composite(im, grad_im)
        except Exception as e:
            Logger.error('Snapshotter: problem compositing ({:})'.format(e))
            self.error('composite error')
            return

        # annotate Draw object with required level of detail
        sf, sf_size, hh = self.small_font(w)
        lf, lf_size, hh = self.large_font(w)
        rowsep = sf_size + 6
        if self.annotation != 'plain':    
            draw = ImageDraw.Draw(im)
            w, h = im.size
            row = self.draw_aligned(draw, 0, 0, w, Component.get('DSO').Name, lf, 
                align='left', rowsep=lf_size+30)

        if self.annotation == 'full':
            self.draw_block(self.DSO_details(), draw, row, 0, w, sf, 
                align='left', rowsep=rowsep)
            self.draw_block(self.processing_details()[::-1], draw, h - 1.5*rowsep, 0, w, sf, 
                align='right', rowsep=-rowsep)
            self.draw_block(self.kit_details(), draw, h - 1.5*rowsep, 0, w, sf, 
                align='left', rowsep=-rowsep) # neg builds from base
            self.draw_block(self.session_details(), draw, 5, 0, w, sf, 
                align='right', rowsep=rowsep)

        # for high density screens, reduce size a little
        im = im.resize((int(im.width/self.imreduction) , int(im.height/self.imreduction)))

        if self.return_image:
            return im

        self.save(im)


    def landscape_view(self):

        # to do: better scaling of fonts based on image size

        # v0.5: ensure tmp is in snapshots dir now that user can start from anywhere
        nm = os.path.join(self.app.get_path('snapshots'), '_tmp.png')
        im = Component.get('View').last_image.copy()
        try:
            imsave(nm, im[::-1])
            orig_im = Image.open(nm)
        except Exception as e:
            self.error('load/save error')
            Logger.error('Snapshotter: load/save problem ({:})'.format(e))
            return

        w, h = orig_im.size
 
        # NB this is indep of screen size, unlike eyepiece view 
        sf, sf_size, small_text_height = self.small_font(w)
        lf, lf_size, large_text_height = self.large_font(w)

        rowgap = small_text_height / 2
        obj_details = self.DSO_details()
        proc_details = self.processing_details()
        sesh_details = self.session_details() + self.kit_details()
        max_rows = max([len(obj_details), len(proc_details), len(sesh_details)])
        total_height = h
        if self.annotation != 'plain':
            total_height = total_height + 2 * rowgap + large_text_height
        if self.annotation == 'full':
            total_height = int(total_height + max_rows * small_text_height + (max_rows - 1) * rowgap)

        im = Image.new('RGBA', (w + 2 * self.framewidth, total_height + 2 * self.framewidth), 
            color=(0, 0, 0) if self.dark_surround else (255, 255, 255))

        im.paste(orig_im, (self.framewidth, self.framewidth))
        draw = ImageDraw.Draw(im)

        w += 2 * self.framewidth
        h += 2 * self.framewidth

        # move row base to base of text
        w += large_text_height

        if self.annotation != 'plain':
            row = self.draw_aligned(draw, h , 0, w, Component.get('DSO').Name, lf, 
                align='center', rowsep=large_text_height + 2 * rowgap)

        if self.annotation == 'full':
            row += rowgap
            draw.line((5, row, w - 5, row), fill=self.grey(), width=1)
            row += (rowgap + small_text_height)
            cols = [int(x) for x in orig_im.size[0] * np.linspace(0, 1, 4)]
            self.draw_block(sesh_details, draw, row,
                cols[0], cols[1], sf, align='left', rowsep=small_text_height + rowgap)
            self.draw_block(obj_details, draw, row, 
                cols[1], cols[2], sf, align='center', rowsep=small_text_height + rowgap)
            self.draw_block(proc_details, draw, row, 
                cols[2], cols[3], sf, align='right', rowsep=small_text_height + rowgap)

        # delete temporary file and save
        os.remove(nm)

        if self.return_image:
            return im
        
        self.save(im)


    def draw_aligned(self, draw, row, l, r, text, font=None, margin=15, rowsep=20, align='center'):
        # helper to draw pixel text aligned; return new row position
        # alignment is within range of cols, a pair (l, r)
        textwidth, textheight = draw.textsize(text, font)
        if align == 'left':
            col = l + margin
        elif align == 'right':
            col = r - textwidth - margin
        elif align == 'center':
            col = (l + r - textwidth) / 2
        draw.text((col, row), text, font=font, fill=self.grey())
        return row + rowsep

    def draw_block(self, block, draw, row, l, r, font, align='center', rowsep=20):
        for v in block:
            if v:
                row = self.draw_aligned(draw, row, l, r, v, font, align=align, rowsep=rowsep)

    def save(self, im):
        # save, handling any format conversions
        try:
            if self.save_format == 'png':
                im.save(self.save_path)
            elif self.save_format == 'jpg':
                bg = Image.new('RGB', im.size, (255, 255, 255))
                bg.paste(im, im)
                bg.save(self.save_path)
            self.info(os.path.basename(self.save_path)[:-4])
        except Exception as e:
            Logger.error('Snapshotter: problem with _save ({:})'.format(e))
            self.error('save error')



''' Handles LRGB and L+narrowband processing
'''

import warnings
warnings.simplefilter('ignore')
import numpy as np
from scipy.stats import trimboth
from skimage.color import rgb2lab, lab2rgb
from skimage.transform import resize, rescale
from kivy.properties import ConfigParserProperty, NumericProperty
from jocular.gradient import estimate_gradient, estimate_background
from jocular.component import Component

def bin_image(im, binfac=2, rescale_to_orig=True):
    ''' bin and rescale to original size; 
        in v0.5 added the preserve range option
    '''
    binfac = int(binfac)
    if binfac == 1:
        return im
    if binfac < 1:
        return im
    im2 = rescale(im, 1 / binfac, anti_aliasing=True, mode='constant', 
        preserve_range=True, multichannel=False)
    if rescale_to_orig:
        return resize(im2, im.shape, anti_aliasing=False, mode='constant',
            preserve_range=True)
    return im2

def _percentile_clip(a, perc=80):
    return np.mean(trimboth(np.sort(a, axis=0), (100 - perc)/100, axis=0), axis=0)

def modify_saturation(x, saturation):
    # Shift endpoints of labA or B to increase saturation. A/B values cover the
    # range -127- to 128 so we increase slope and then trim to range.
    # sat=1 corresponds to shift of 100 (that's a lot!)
    if saturation <= 0:
        return x
    k = saturation * 100
    mx = np.mean(x)
    # steepen but preserve mean (perhaps should do this after trimming?)
    x1 = (x - mx) * (255/(255 - 2*k)) + mx 
    x1[x1 < -127] = -127
    x1[x1 > 127] = 127
    return x1

def modify_hue(x, shift):
    # Shift mean of x by given amount
    x = x + shift
    return x

class MultiSpectral(Component):

    colour_binning = ConfigParserProperty(
        1, 'Colour', 'colour_binning', 'app', val_type=int)
    subtract_colour_gradients = ConfigParserProperty(
        1, 'Colour', 'subtract_colour_gradients', 'app', val_type=int)
    
    saturation = NumericProperty(0)
    colour_stretch = NumericProperty(.5)
    redgreen = NumericProperty(0)
    yellowblue = NumericProperty(0)

    def on_new_object(self):
        self.reset()

    def reset(self):
        # set these to None to force updates in the colour-generation sequence
        self.lum = None
        self.R = None
        self.G = None
        self.B = None
        self.normed_RGB = None
        self.LAB = None          
        self.A_sat = None
        self.B_sat = None
        self.A_hue = None
        self.B_hue = None
        self.RGB = None
        self.layer = None
        self.layer_stretched = None

    def on_saturation(self, *args):
        self.adjust_saturation()

    def on_colour_stretch(self, *args):
        self.adjust_colour_stretch()

    def on_redgreen(self, *args):
        self.adjust_hue()

    def on_yellowblue(self, *args):
        self.adjust_hue()

    def LRGB_changed(self, L=None, R=None, G=None, B=None):

        self.reset()
        # we lack one of RGB, so let Monochrome handle it and call reset
        # so colour components are regenerated next time we have RGB
        if R is None or G is None or B is None:
            self.luminance_only()
        else: 
            # we have RGB; if no L, generate synthetic lum
            if L is None:
                L = .2125*R  + .7154*G + .0721*B
                L = L / np.percentile(L.ravel(), 99.99)
                 
            self.lum = Component.get('Monochrome').update_lum(L)
            self.R = R
            self.G = G
            self.B = B
            self.create_LAB()

    def L_plus_changed(self, L=None, layer=None):

        # if we have no layer, update luminance only
        self.reset() # ensure components get updated
        if layer is None:
            self.luminance_only()
        elif L is not None:
            self.lum = Component.get('Monochrome').update_lum(L)
            self.layer = layer / np.percentile(layer.ravel(), 99.99)
            self.layer_stretch_changed()

    def layer_stretch_changed(self):
        if self.layer is not None:
            self.layer_stretched = self.modgamma(self.layer - estimate_gradient(self.layer)[0], 
                g=self.colour_stretch)
            self.layer_sat_changed()

    def layer_sat_changed(self):
        if self.layer_stretched is not None:
            r = 3 * self.saturation * self.layer_stretched + self.lum
            r[r < 0] = 0
            r[r > 1] = 1            
            Component.get('View').display_image(np.stack([r, self.lum, self.lum], axis=-1))

    def luminance_only(self):
        Component.get('Monochrome').L_changed(
            Component.get('Stacker').get_stack(filt='all'))

    def luminance_updated(self, lum):
        # called from monochrome if user updates luminance control e.g. white

        self.lum = lum
        # we are in LRGB mode
        if self.LAB is not None:
            self.create_RGB()

        # we are in L+ mode
        elif self.layer is not None:
            self.layer_sat_changed()

        # display as mono
        else:            
            Component.get('View').display_image(lum)

    def on_colour_binning(self, *args):
        self.create_LAB()

    def on_subtract_colour_gradients(self, *args):
        self.create_LAB()

    ''' Colour processes are daisy-chained and intermediate representations cached for speed        
        create_LAB          normed_RGB
        stretch_color       LAB             
        adjust_saturation   A_sat, B_sat
        adjust_hue          A_hue, B_hue
        create_RGB          RGB             <-- entry point for luminosity changes
    '''

    def create_LAB(self):
        # Initial stage of LAB image creation from RGB stacks

        # def bin_image(im, binfac=2):
        #     im2 = rescale(im, 1 / binfac, anti_aliasing=True, mode='constant', multichannel=False)
        #     return resize(im2, im.shape, anti_aliasing=False, mode='constant')

        def _limit(im):
            im[im < 0] = 0
            im[im > 1] = 1
            return im

        if not self.avail(['R', 'G', 'B']):
            return

        ims = [self.R, self.G, self.B]

        # binning ~ 70 ms
        if self.colour_binning:
            ims = [bin_image(im) for im in ims]

        # alternative approach to test
        # might also add option to not subtract anything.... just to test
        if self.subtract_colour_gradients:
            ims = [im - estimate_gradient(im) for im in ims]
        else:
            ims = [im - estimate_background(im)[0] for im in ims]

        # it turns out this value is absolutely critical because setting it too low means that stars get saturated
        # whereas too high, or just using the absolute max leads to a colour space is too compressed (too grey).

        max_pixel_vals = [np.percentile(im.ravel(), 99.99) for im in ims]
        ims = [im / max(max_pixel_vals) for im in ims]

        # now limit image to 0-1 because we are about to stretch
        self.normed_RGB = [_limit(im) for im in ims]

        # rest of process
        self.adjust_colour_stretch()

    def modgamma(self, x, g=.5, a0=.01):
        y = x.copy()
        s = g / (a0 * (g - 1) + a0 ** (1 - g))
        d = (1 / (a0 ** g * (g - 1) + 1)) - 1
        y[x < a0] = x[x < a0] * s
        y[x >= a0] =  (1 + d) * (x[x >= a0] ** g) - d
        return y

    def avail(self, props):
        if type(props) != list:
            props = [props]
        return all([hasattr(self, p) and getattr(self, p) is not None for p in props])

    def adjust_colour_stretch(self, *args):
        # intercept LAB process here if we need to stretch colour

        if self.avail('layer'):
            self.layer_stretch_changed()

        elif self.avail('normed_RGB'):
            # use modified gamma correction here to avoid colour noise
            ims = [self.modgamma(im, g=self.colour_stretch) for im in self.normed_RGB]
            self.LAB = rgb2lab(np.stack(ims, axis=-1))  # 80ms
            self.adjust_saturation()

    def adjust_saturation(self, *args):
        # Modify saturation and call next process
        
        if self.avail('layer'):
            self.layer_sat_changed()

        elif self.avail('LAB'):
            self.A_sat = modify_saturation(self.LAB[:, :, 1], self.saturation)
            self.B_sat = modify_saturation(self.LAB[:, :, 2], self.saturation)
            self.adjust_hue() # next process in hue adjustment

    def adjust_hue(self, *args):
        # Modify hue and generate RGB image
        
        if self.avail('A_sat'):
            self.A_hue = modify_hue(self.A_sat, self.redgreen)
            self.B_hue = modify_hue(self.B_sat, self.yellowblue)
            self.create_RGB()  # next process

    def create_RGB(self):
        ''' Combine L, A and B into RGB image. Called here when previous colour process
        has been applied (hue_changed), or when luminosity component has changed (below)
        '''
        if self.avail(['lum', 'A_hue']):
            self.RGB = lab2rgb(np.stack([ 100 * self.lum, self.A_hue, self.B_hue], axis=-1))
            Component.get('View').display_image(self.RGB)




''' View: represents the actual image on the screen
'''

import os
import time
import numpy as np
import math

from kivy.clock import Clock
from kivy.app import App
from kivy.lang import Builder
from kivy.uix.scatter import Scatter
from kivy.graphics.texture import Texture
from kivy.graphics.transformation import Matrix
from kivy.properties import (
    BooleanProperty, 
    NumericProperty, 
    BoundedNumericProperty, 
    ConfigParserProperty
)
from kivy.core.window import Window
from kivy.base import stopTouchApp
from kivy.animation import Animation

from jocular.utils import angle360
from jocular.component import Component
from jocular.metrics import Metrics
from jocular.widgets import JBubble

Builder.load_string('''
<View>:
    size: image.size
    size_hint: None, None 
    auto_bring_to_front: False
    do_rotation: True
    do_translation: True
    do_scale: True
    Image:
        id: image
''')

class View(Scatter, Component):

    zoom = NumericProperty(0)  # zoom lever ranges from 0 to 1 and maps from min-max zoom using curve
    show_help = BooleanProperty(False)
    invert = BooleanProperty(False)
    orientation = BoundedNumericProperty(0, min=0, max=360)
    show_reticle = BooleanProperty(False)
    brightness = NumericProperty(1)
    is_dimmed = BooleanProperty(False)
    transparency = NumericProperty(0)
    flip_UD = ConfigParserProperty(0, 'Flipping', 'flip_UD', 'app', val_type=int)
    flip_LR = ConfigParserProperty(0, 'Flipping', 'flip_LR', 'app', val_type=int)
    min_zoom = ConfigParserProperty(.5, 'Zoom', 'min_zoom', 'app', val_type=float)
    max_zoom = ConfigParserProperty(30, 'Zoom', 'max_zoom', 'app', val_type=float)
    zoom_power = ConfigParserProperty(1, 'Zoom', 'zoom_power', 'app', val_type=float)
    confirm_on_close = ConfigParserProperty(0, 'Confirmations', 'confirm_on_close', 'app', val_type=int)
    dim_screen_after = ConfigParserProperty(5, 'Dimming', 'dim_screen_after', 'app', val_type=int)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.cached_image = None
        self.last_image_time = 0
        self.app = App.get_running_app()
        # self.last_image_dir = 'lastimage'
        # add_if_not_exists(self.last_image_dir) # need to fix this
        self.app.gui.add_widget(self, index=100)
        self._set_center((0, 0))
        self.ring_selected = False
        self.image_selected = False
        self.old_transparency = self.transparency
        self.old_brightness = self.brightness

        # start listening to mouse position events etc
        Window.bind(mouse_pos=self.on_mouse_pos)

    def on_mouse_pos(self, *args):
        if hasattr(self, 'dimmer_event'):
            self.dimmer_event.cancel()
        if self.is_dimmed:
            Animation(brightness=self.old_brightness, duration=.2).start(self)
            Animation(transparency=self.old_transparency, duration=.2).start(self)
            self.is_dimmed = False
        if self.dim_screen_after > 0:
            self.dimmer_event = Clock.schedule_once(self.dim_screen, self.dim_screen_after)

    def dim_screen(self, dt):
        if self.last_image is None:
            return
        self.old_transparency = self.transparency
        self.old_brightness = self.brightness
        Animation(brightness=0, duration=.5).start(self)
        Animation(transparency=1, duration=.5).start(self)
        self.is_dimmed = True

    def reset(self):
        # Keep as separate method because most likely called on stack reset
        self.last_image = None
        self.reset_texture()

    def on_new_object(self):
        # until we receive the new object' sizes we cannot really zoom, so do this on first draw
        self.reset()
        # preserve orientation for new objects
        if Component.get('ObjectIO').existing_object:
            self.settings = Component.get('Metadata').get({'orientation'})

    def on_save_object(self):
        Component.get('Metadata').set('orientation', self.orientation)

    def on_show_reticle(self, *args):
        self.app.gui.gui['reticle']['widget'].show = self.show_reticle

    def on_brightness(self, *args):
        self.app.brightness = self.brightness

    def on_transparency(self, *args):
        self.app.transparency = max(0, self.transparency)

    def update_state(self):
        s = 'L-R, ' if self.flip_LR else ''
        s += 'U-D' if self.flip_UD else ''
        if s.endswith(', '):
            s = s[:-2]
        flipstr = ' | flip {:}'.format(s) if s else ''
        invstr = ' | inv' if self.invert else ''
        self.info('rot {:.0f}\u00b0 | {:4.1f}x{:}{:}'.format(
            self.orientation, self.lever_to_zoom(self.zoom), flipstr, invstr))
        Component.get('Annotator').update()

    def lever_to_zoom(self, z):
        return self.min_zoom + (z ** self.zoom_power) *(self.max_zoom - self.min_zoom)

    def zoom_to_lever(self, z):
        if z < self.min_zoom: 
            return 0
        return ((z - self.min_zoom) / (self.max_zoom - self.min_zoom)) ** (1 / self.zoom_power)

    def on_zoom(self, *args):
        # zoom anchored on window centre
        scale = self.lever_to_zoom(self.zoom) / self.scale
        mat = Matrix().scale(scale, scale, scale)
        self.apply_transform(mat, anchor=Metrics.get('origin'))
        self.update_state()

    # eventually find a better place for these
    def confirm_close(self, *args):
        if self.confirm_on_close:
            JBubble(actions={'Really close?': self.close}, loc='mouse').open()
        else:
            self.close()

    def close(self, *args):
        stopTouchApp()

    def open_settings(self, *args):
        self.app.open_settings()
 
    def on_orientation(self, *args):
        # rotate anchored on window center
        rot = self.orientation - self._get_rotation()
        self.apply_transform(Matrix().rotate((np.pi/180)*rot, 0 ,0, 1), anchor=Metrics.get('origin'))
        self.update_state()

    def fit_to_window(self, *args):
        # compute zoom to fit to window
        if self.last_image is None:
            return
        shape = self.last_image.shape
        h, w = shape[0], shape[1]  # done like this because might be 3-D
        if  w < 101:
            return
        z = self.zoom_to_lever(Window.height / h)
        App.get_running_app().gui.set('zoom', z)
        self.zoom = z
        self.orientation = 0
        self._set_center(Metrics.get('origin'))

    def on_invert(self, *args):
        if self.last_image is None:
            return
        # only invert if not colour image    
        if self.ids.image.texture.colorfmt == 'luminance':
            self.display_image(use_cached_image=True)
            self.update_state()
        elif self.invert:
            self.invert = False

    def on_flip_LR(self, *args):
        self.display_image(use_cached_image=True)
        self.update_state()

    def on_flip_UD(self, *args):
        self.display_image(use_cached_image=True)
        self.update_state()

    def reset_texture(self, w=10, h=10, colorfmt='luminance'):
        self.ids.image.size = h, w
        self.ids.image.texture = Texture.create(size=(h, w), colorfmt=colorfmt)
        self._set_center(Metrics.get('origin'))
        self.w, self.h = w, h
        self.colorfmt = colorfmt

    def display_blank(self):
        self.display_image(0 * np.ones((self.w, self.h)))

    def do_flips(self, im):
        # perform flips (< 1 ms)
        if self.flip_UD:
            im = im[::-1]
        if self.flip_LR:
            im = np.fliplr(im)
        return im

    def display_image(self, im=None, use_cached_image=False):
        ''' Called with an image, in which case update cached image, perform flips etc
        and display; or without an image, in which case use cached image and perform
        flips/invert directly on that.
        '''

        if (im is None) and (not use_cached_image):
            return

        if use_cached_image and (self.cached_image is None):
            return

        if im is not None:
            self.cached_image = im

        if use_cached_image:
            im = self.cached_image  # just for shorthand below

        # we now have a cached image
        colorfmt = 'luminance' if im.ndim == 2 else 'rgb'

        # check if shape or color has changed
        w, h = im.shape[0], im.shape[1]
        if (w != self.w) or (h != self.h) or (colorfmt != self.colorfmt):
            # for some reason, image is recentered after texture reset but orientation/size not affected 
            old_center = self._get_center()
            self.reset_texture(w=w, h=h, colorfmt=colorfmt)
            self._set_center(old_center)

        im = self.do_flips(im)

        # only invert if luminance and not RGB image
        if self.invert and colorfmt == 'luminance':
            im = np.subtract(1, im, dtype=im.dtype)

        self.last_image = np.uint8(im * 255)
        self.ids.image.texture.blit_buffer(self.last_image.flatten(), colorfmt=colorfmt, bufferfmt='ubyte')

        # first time through for each object, apply settings
        if hasattr(self, 'settings') and self.settings is not None:
            z = self.zoom_to_lever(Window.height / self.last_image.shape[0])
            App.get_running_app().gui.set('zoom', z)
            self.zoom = z
            if 'orientation' in self.settings:
                self.orientation = self.settings['orientation']
            self._set_center(Metrics.get('origin'))
            self.invert = False
            self.settings = None  # indicates that they have been applied

        # refresh image for monitor at most every 300 ms
        # currently disabled
        if False:
            if (time.time() - self.last_image_time) > .3:
                np.save(os.path.join(
                    self.last_image_dir, 'im_{:}.npy'.format(time.time())), self.last_image)
                self.last_image_time = time.time()


    # handle mouse events for rotation/translation

    def on_touch_down(self, touch):
        # check if touch is within eyepiece and if so, where

        x, y = touch.pos
        xc, yc = Metrics.get('origin')
        r = (((xc - x) ** 2 + (yc - y) ** 2) ** .5) / Metrics.get('inner_radius')

        # touch is not in the eyepiece
        if r  >  .98:
            return False

        # in image
        in_image = self.collide_point(*touch.pos)

        # is it in the outer zone of the image, or not touching image, rotate
        if not in_image or r > .8:
            self.ring_selected = True
            self.theta = math.atan2(y - yc, x - xc) / (math.pi/ 180)
            self.xc = xc
            self.yc = yc
            return True

        if in_image:
            if touch.is_double_tap:
                self.invert = not self.invert
                return True
            self.image_selected  = True
            self.last_x = x
            self.last_y = y
            return True

        return False

    def on_touch_move(self, touch):
        if self.ring_selected:
            x, y = touch.pos
            theta = math.atan2(y - self.yc, x - self.xc) / (math.pi / 180)
            self.orientation = angle360(self.orientation + (theta - self.theta))
            self.theta = theta
            return True
        if self.image_selected:
            x, y = touch.pos
            self.x += (x - self.last_x)
            self.y += (y - self.last_y)
            self.last_x = x
            self.last_y = y
            Component.get('Annotator').update()
            return True
        return False

    def on_touch_up(self, touch):
        self.ring_selected = False
        self.image_selected = False



''' Handles calibration library and actual calibration of subs.
'''

import os.path
import shutil
import numpy as np
import time
from datetime import datetime
from scipy.stats import trimboth

from kivy.app import App
from kivy.logger import Logger
from kivy.properties import BooleanProperty, DictProperty, ConfigParserProperty
from kivy.core.window import Window

from jocular.table import Table
from jocular.utils import add_if_not_exists, make_unique_filename
from jocular.component import Component
from jocular.image import Image, save_image, fits_in_dir

date_time_format = '%d %b %y %H:%M'

class Calibrator(Component):

    masters = DictProperty({})
    apply_flat = BooleanProperty(False)
    apply_dark = BooleanProperty(False)
    apply_bias = BooleanProperty(False)
    use_l_filter = ConfigParserProperty(1, 'Flats', 'use_l_filter', 'app', val_type=int)
    temperature_tol = ConfigParserProperty(5, 'Darks', 'temperature_tol', 'app', val_type=float)
    exposure_tol = ConfigParserProperty(5, 'Darks', 'exposure_tol', 'app', val_type=float)
    dark_days_tol = ConfigParserProperty(1, 'Darks', 'dark_days_tol', 'app', val_type=float)
    flat_days_tol = ConfigParserProperty(30, 'Flats', 'flat_days_tol', 'app', val_type=float)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        self.calibration_dir = self.app.get_path('calibration')
        self.load_masters()

    def new_master_from_watcher(self, m):
        # construct unique/sensible name
        dest = make_unique_filename(os.path.join(self.calibration_dir, m.name))
        try:
            os.rename(m.path, dest)
            m.path = dest
            m.name = os.path.basename(dest)
            self.add_to_library(m)
        except Exception as e:
            raise IOError('cannot move {:} ({:})'.format(m.path, e))
            # Logger.error('Calibrator: cannot move {:} ({:})'.format(m.path, e))

    def load_masters(self):
        # construct metadata by reading and validating FITs in calibration dir
        self.masters = {}
        self.library = {}
        for f in fits_in_dir(self.calibration_dir):
            path = os.path.join(self.calibration_dir, f)
            try:
                s = Image(path)
                if s.is_master:
                    self.add_to_library(s)
            except Exception as e:
                Logger.warn('Calibrator: unable to parse calibration {:} ({:})'.format(f, e))

    def add_to_library(self, m):
        # m is a validated masters, so just pick off required fields
        name = m.name
        self.masters[name] = m
        self.library[name] = {
            'name': name,
            'type': m.sub_type,
            'exposure': str(m.exposure) if m.exposure is not None else '???',
            'temperature': str(m.temperature) if m.temperature is not None else '???',
            'filter': m.filter,
            'created': m.create_time.strftime(date_time_format),
            'shape_str': m.shape_str,
            'age': m.age,
            'nsubs': m.nsubs if m.nsubs is not None else 0
        }


    def save_master(self, sub_type=None, exposure=None, temperature=None, filt=None):
        # Save master from existing stack, applying any required post-processing
        # masters are saved to watched directory so that processing is uniform re external captures

        # generate master from stack
        stacker = Component.get('Stacker')

        # force the use of method that the user has chosen or set up by default for this type of calib
        master = stacker.get_stack(filt, calibration=True)

        # apply bad pixel mapping to calibration frames
        # if dark, find hot pixels in master and remove, otherwise use existing BPM
        # not yet tested
        bpm = Component.get('BadPixelMap')
        if sub_type == 'dark':
            master = bpm.do_bpm(master, bpm.find_hot_pixels(master))
        else:
            master = bpm.do_bpm(master)

        # flats were divided thru by their robust mean to account for level differences 
        # but then scaled to 50% to enable B/W controls; so multiply by 2
        if sub_type == 'flat':
            master = 2 * master

        # save calibration master to Watched so that it is processed in the same way as when user drops master)
        save_image(data=master,
            path=os.path.join(self.app.get_path('watched'), 'master{:}.fit'.format(sub_type)),
            exposure=exposure,
            filt=filt,
            temperature=temperature,
            sub_type='master ' + sub_type,
            nsubs=stacker.get_selected_sub_count())

        # give it time to save (doesn't seem to help)
        # time.sleep(1)

        # add to notes field of current DSO
        Component.get('Notes').notes = 'exposure {:} filter {:} temperature {:}'.format(exposure, filt, temperature)

    def calibrate(self, sub):
        # Given a light sub, apply calibration. Fails silently if no suitable calibration masters. 

        sub.calibrations = set({})

        if not self.library: 
            return

        # get all masters (check speed, but should be quick)
        dark = self.get_dark(sub)
        flat = self.get_flat(sub)
        bias = self.get_bias(sub)

        # Logger.debug('Calibrator: D {:} F {:} B {:}'.format(dark, flat, bias))

        D = self.get_master(dark)
        # if D is not None:
        #     print('{:} min {:} max {:} median {:} mean {:}'.format(dark, np.min(D), np.max(D), np.median(D), np.mean(D)))
        F = self.get_master(flat)
        # if F is not None:
        #     print('{:} min {:} max {:} median {:} mean {:}'.format(flat, np.min(F), np.max(F), np.median(F), np.mean(F)))
        B = self.get_master(bias)
        # if B is not None:
        #     print('{:} min {:} max {:} median {:} mean {:}'.format(bias, np.min(B), np.max(B), np.median(B), np.mean(B)))

        im = sub.get_image()
        if self.apply_dark and self.apply_flat:
            if dark is not None and flat is not None:
                im = (im - D) / F
                sub.calibrations = {'dark', 'flat'}
            elif dark is not None:
                im = im - D
                sub.calibrations = {'dark'}
            elif flat is not None:
                if bias is not None:
                    sub.calibrations = {'flat', 'bias'}
                    im = (im - B) / F
                else:
                    sub.calibrations = {'flat'}
                    im = im / F # inadvisable, but we allow it

        elif self.apply_dark:
            if dark is not None:
                im = im - D
                sub.calibrations = {'dark'}

        elif self.apply_flat:
            if flat is not None:
                if bias is not None:
                    sub.calibrations = {'flat', 'bias'}
                    im = (im - B) / F
                else:
                    sub.calibrations = {'flat'}
                    im = im / F

        elif self.apply_bias:
            if bias is not None:
                sub.calibrations = {'bias'}
                im = im - B

        # limit
        im[im < 0] = 0
        im[im > 1] = 1

        sub.image = im
        applied = ' '.join(list(sub.calibrations))
        if applied:
            self.info('applied ' + applied)
        else:
            self.info('none')

    def get_dark(self, sub):
        # Find suitable dark for this sub given its parameters

        if sub.exposure is None:
            return None

        # choose darks that are the right shape with exposure within tolerance
        darks = {k: v for k, v in self.masters.items()
                    if  v.shape == sub.shape and 
                        v.sub_type == 'dark' and 
                        v.exposure is not None and 
                        abs(v.exposure - sub.exposure) < self.exposure_tol}

        temperature = Component.get('Session').temperature

        if temperature is not None:
            # we know temperature, select those with temperatures and within tolerance
            darks = [k for k, v in darks.items() if 
                v.temperature is not None and abs(v.temperature - temperature) < self.temperature_tol]
        else:
            # find those within date tolerance (set to 1 to get darks in current session)
            darks = [k for k, v in darks.items() if v.age < self.dark_days_tol]

        # if we have darks, return name of first one
        return darks[0] if len(darks) > 0 else None


    def get_bias(self, sub):
        # get the most recent bias

        bias = {k: v.age for k, v in self.masters.items() 
                if v.shape == sub.shape and v.sub_type == 'bias' }

        return min(bias, key=bias.get) if len(bias) > 0 else None
 
    def get_flat(self, sub):

        # flats of right shape
        flats = {k:v for k, v in self.masters.items() 
            if  v.shape == sub.shape and v.sub_type == 'flat'}

        # flat in required filter
        if sub.filter is not None:
            flats_in_filt = {k: v for k, v in flats.items() if v.filter is not None and v.filter == sub.filter}
        else:
            flats_in_filt = {} 

        # if we have none and can use L filter, use these
        if (len(flats_in_filt) == 0) and self.use_l_filter:
            flats_in_filt = {k:v for k, v in flats.items() if v.filter == 'L'}

        # do we have any now? if not, return
        if len(flats_in_filt) == 0:
            return None

        # find any within day tolerance, noting that this compares the date of the flat with
        # the date of the sub (i.e. not necessarily the current date)
        flats = {k: abs(v.create_time - sub.create_time).days for k,v in flats_in_filt.items()}
        flats = {k: v for k, v in flats.items() if v <= self.flat_days_tol}

        # find most recent if there is a choice
        for k in sorted(flats, key=flats.get):
            return k
        return None

    def get_master(self, name):
        if name is None:
            return None
        # Retrieve image (NB loaded on demand, so effectively a cache)
        # return self.masters[name].get_master_data()
        # changed in v0.4.5 because getting image is now same regardless of sub or master
        return self.masters[name].get_image()

    def _most_subs(self, cands):
        c = {k: cands[k]['nsubs'] for k in cands.keys()}
        return max(c, key=c.get)

    def calibrate_flat(self, sub):
        ''' Perform calibrations on flat which include subtracting bias if
        available , and rescaling so the mean intensity is .5 (because outlier rejection 
        methods used to combine flat subs work best with normalised frames due to changing 
        light levels; the value of .5 is so that we can use B & W controls; we rescale to 
        a mean of 1 when saving since this is what a good flat needs for dividing)
        '''

        im = sub.get_image()
        #print('before calibration min {:} max {:} median {:}'.format(
        #    np.min(im), np.max(im), np.median(im)))

        # subtract bias if available
        bias = self.get_bias(sub)
        if bias is not None:
            #print('subtracting bias')
            im = im - self.get_master(bias)

        #print('after bias min {:} max {:} median {:}'.format(
        #    np.min(im), np.max(im), np.median(im)))

        # normalise by mean of image in central 3rd zone 
        perc = 75  # retain central 75% of points when computing mean 
        w, h = im.shape
        w1, w2 = int(w / 3), int(2 * w / 3)
        h1, h2 = int(h / 3), int(2 * h / 3)
        imr = im[h1: h2, w1: w2]
        # imr = im[(h // 2 - r):(h // 2 + r), (w // 2 - r):(w // 2 + r)]
        robust_mean = np.mean(trimboth(np.sort(imr.ravel(), axis=0), 
            (100 - perc)/100, axis=0), axis=0)
        #print('robust mean in w {:}-{:}, h {:}-{:} is {:}'.format(
        #    w1, w2, h1, h2, robust_mean))

        sub.image = .5 * im / robust_mean
        #print('after bias min {:} max {:} median {:}'.format(
        #    np.min(sub.image), np.max(sub.image), np.median(sub.image)))


    # calibration table handling ---------------------------------------------------------------------

    def build_calibrations(self):

        return Table(
            size=Window.size,
            data=self.library,
            name='Calibration masters',
            description='Calibration masters',
            cols={
                'Name': {'w': 300, 'align': 'left', 'field': 'name'},
                'Type': {'w': 60, 'field': 'type', 'align': 'left'},
                'Exposure': {'w': 80, 'field': 'exposure'},
                'Temp. C': {'w': 80, 'field': 'temperature', 'type': str},
                'Filter': {'w': 80, 'field': 'filter'},
                'Created': {'w': 180, 'field': 'created', 'sort': {'DateFormat': date_time_format}},
                'Size': {'w': 70, 'field': 'shape_str'},
                'Age': {'w': 50, 'field': 'age', 'type': int},
                'Subs': {'w': 50, 'field': 'nsubs', 'type': int}
                },
            actions={'move to delete dir': self.move_to_delete_folder},
            on_hide_method=self.app.table_hiding
            )

    def show_calibration_table(self, *args):
        '''Called from menu'''

        if not hasattr(self, 'calibration_table'):
            self.calibration_table = self.build_calibrations()
        self.app.showing = 'calibration'

        # check for redraw
        if self.calibration_table not in self.app.gui.children:
            self.app.gui.add_widget(self.calibration_table, index=0)

        self.calibration_table.show()    

    def move_to_delete_folder(self, *args):
        add_if_not_exists('deleted')
        for nm in self.calibration_table.selected:
            if nm in self.library:
                to_path = os.path.join('deleted', nm + datetime.now().strftime('%d_%b_%y_%H_%M_%S'))
                from_path = os.path.join(self.calibration_dir, nm)
                try:
                    shutil.move(from_path, to_path)
                    del self.library[nm]
                    del self.masters[nm]
                except Exception as e:
                    self.error('problem moving master')
                    Logger.error('Calibrator: problem moving master to {:} ({:})'.format(to_path, e))
        self.calibration_table.update()




''' Filterwheel: represents collection of filters available as well 
    as providing control of an electronic filter wheel (EFW) where 
    available. All the individual filter information should reside 
    here e.g. on-screen colour representations, transmissibilities 
    if needed for ordering taking multiple flats.

    Currently supports (hard-coded) the SX USB filter wheel only.
'''

# import hid   # human interface device, to allow access to SX EFW
from kivy.app import App
from kivy.logger import Logger
from kivy.clock import Clock
from kivy.properties import StringProperty
from jocular.component import Component
from jocular.widgets import JPopup


class FilterWheel(Component):

    filter_properties = {
        "L": {"color": (0.7, 0.7, 0.7), "bg_color": (0.1, 0.1, 0.1, 0.5)},
        "R": {"color": (1, 0, 0), "bg_color": (1, 0, 0, 0.6)},
        "G": {"color": (0, 1, 0), "bg_color": (0, 1, 0, 0.5)},
        "B": {"color": (0, 0, 1), "bg_color": (0, 0, 1, 0.5)},
        "dark": {"color": (0.1, 0.1, 0.1), "bg_color": (0, 0, 0, 0)},
        "Ha": {"color": (0.8, 0.3, 0.3), "bg_color": (1, 0.6, 0.6, 0.5)},
        "OIII": {"color": (0.2, 0.6, 0.8), "bg_color": (0, 0, 0, 0)},
        "SII": {"color": (0.1, 0.8, 0.4), "bg_color": (0, 0, 0, 0)},
        "spec": {"color": (0.7, 0.6, 0.2), "bg_color": (0, 0, 0, 0)},
        "-": {"color": (0.3, 0.3, 0.3), "bg_color": (0, 0, 0, 0)},
    }

    current_filter = StringProperty("L")

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # form array of filters
        config = App.get_running_app().config
        self.filters = [
            config.get("Filters", "filter{:}".format(i + 1)) for i in range(9)
        ]

    def on_new_object(self):
        self.connected = False
        self.connect()
        if self.connected:
            self.info("connected")
        else:
            self.info("not connected")

    def on_current_filter(self, *args):
        App.get_running_app().gui.set("filter_button", self.current_filter)
        self.info("{:} filter".format(self.current_filter))

    def update_filter(self, fnum, ftype):
        # called when config changes in jocular itself
        if fnum.startswith("filter"):
            findex = int(fnum[6:]) - 1
            self.filters[findex] = ftype

    # methods for other components to access things like filter colours
    def get_color(self, name="L"):
        if name in self.filter_properties:
            return self.filter_properties[name]["color"]
        else:
            # mid-grey
            if name != "empty":
                self.warn("no filter called {:}".format(name))
            return 0.5, 0.5, 0.5, 1

    def on_close(self):
        if self.connected and self.efw:
            self.efw.close()

    def connect(self):
        # Try to connect

        if self.connected:
            return

        self.connected = False

        try:
            import hid
        except Exception as e:
            Logger.warn("FilterWheel: cannot import HID ({:})".format(e))
            return
        try:
            if hasattr(self, "efw"):
                if self.efw:
                    self.efw.close()
            else:
                self.efw = hid.device()
        except Exception as e:
            Logger.warn("FilterWheel: failed to get HID device ({:})".format(e))
            return

        # note that if we open when already open, it fails
        try:
            self.efw.open(0x1278, 0x0920)
            self.connected = True
            Logger.info("FilterWheel: connected!")
        except Exception as e:
            Logger.trace("FilterWheel: failed to open device ({:})".format(e))
            self.connected = False

    def order_by_transmission(self, filts):
        #  need to add others to this and perhaps do it via filter_properties
        if type(filts) != list:
            filts = [filts]
        t_order = ["Ha", "B", "G", "R", "L"]
        return [f for f in t_order if f in filts]

    #  since filter change takes some seconds, we use callbacks to control actions
    #  once the filterwheel is in position; in simple cases where there are no filters
    #  or user is requesting same filter, no need for callbacks -- up to the caller

    def select_filter(self, name="L", changed_action=None, not_changed_action=None):
        """Connect to filter by name if possible, calling appropriate callback if provided"""

        Logger.info("FilterWheel: select filter {:}".format(name))

        # no change of filter so execute callback if it exists, or return otherwise
        if self.current_filter == name:
            if changed_action is not None:
                changed_action()
            return

        # we need to change filter: does the filter exist?
        if (name not in {"dark", "L"}) and name not in self.filters:
            self.warn("Cannot find filter {:}".format(name))
            if not_changed_action is not None:
                not_changed_action()
            return

        # if we have a filterwheel, then execute the change
        if self.connected:
            position = self.filters.index(name) + 1  # 1-indexed
            try:
                self.efw.write([position + 128, 0])
                if changed_action is not None:
                    Clock.schedule_once(changed_action, 3)
                self.current_filter = name
            except Exception as e:
                Logger.error(
                    "FilterWheel: exception moving filterwheel ({:})".format(e)
                )
                self.warn("problem moving EFW")
                if not_changed_action is not None:
                    not_changed_action()
        # prompt user
        else:
            if name == "dark":
                title = "Cap scope for darks"
            else:
                title = "Change filter to {:}".format(name)
            self.current_filter = name
            JPopup(title=title, actions={"Done": changed_action}).open()



''' Handle info.json in its various instantiations over the different versions
'''

import os
import json
import platform
from datetime import datetime
from kivy.logger import Logger
from jocular.image import fits_in_dir
from jocular.component import Component


def remove_empties(d):
    return {
        k: v
        for k, v in d.items()
        if not ((v == '') or (v is None) or (v == {}) or (v == []))
    }


def get_metadata(path, simple=False):
    # Read metadata from path, constructing if necessary
    #  supports info.json (v1/2 of Jocular) and info3.json (v3)
    #  main difference is that info3 is simpler; will always read info3 in pref if both exist

    v1 = False
    try:
        # if we can find v3 metadata, load it
        with open(os.path.join(path, 'info3.json'), 'r') as f:
            md = json.load(f)
    except:
        try:
            #  if pre v3 metadata exists, load it
            with open(os.path.join(path, 'info.json'), 'r') as f:
                md = json.load(f)
            # convert
            newmd = {}
            for p in [
                'Name',
                'Con',
                'OT',
                'Notes',
                'session_notes',
                'SQM',
                'temperature',
                'rejected',
                'seeing',
                'transparency',
                'telescope',
                'camera',
                'exposure',
                'sub_type',
            ]:
                if p in md:
                    newmd[p] = md[p]
            if 'scope' in md and 'orientation' in md['scope']:
                newmd['orientation'] = md['scope']['orientation']
            md = newmd
            v1 = True
        except:
            # cannot find any metadata, so set to empty
            md = {}

    #  if no name, use name of directory contaiining FITs
    if 'Name' not in md:
        md['Name'] = os.path.basename(path)

    md = remove_empties(md)

    # create an infov3 for speed of later loading all obs
    if v1:
        try:
            with open(os.path.join(path, 'info3.json'), 'w') as f:
                json.dump(md, f, indent=1)
        except:
            pass

    #  compute session date/time and number of subs dynamically
    fits = fits_in_dir(path)
    md['nsubs'] = len(fits)
    if len(fits) > 0:
        # pre v0.4.5 approach
        md['session'] = datetime.fromtimestamp(os.path.getmtime(fits[0])).strftime(
            '%d %b %y %H:%M'

        # this one doesn't work on Windows it seems: the create date is wrong
        # md['session'] = datetime.fromtimestamp(creation_date(fits[0])).strftime(
        #    '%d %b %y %H:%M'
        )

    # much of the time we just want metadata for observing list in
    # simple format
    if simple:
        return { 
            'Name': md.get('Name', ''),
            'OT': md.get('OT', ''),
            'Con': md.get('Con', ''),
            'Session': md.get('session', ''),
            'N': md.get('nsubs', 0),
            'Notes': md.get('Notes', '')
            }

    return md


def creation_date(path_to_file):
    ''' Code from 
        https://stackoverflow.com/questions/237079/how-to-get-file-creation-modification-date-times-in-python
        Try to get the date that a file was created, falling back to when it was
        last modified if that isn't possible.
        See http://stackoverflow.com/a/39501288/1709587 for explanation.
    '''
    if platform.system() == 'Windows':
        return os.path.getctime(path_to_file)
    else:
        stat = os.stat(path_to_file)
        try:
            return stat.st_birthtime
        except AttributeError:
            # We're probably on Linux. No easy way to get creation dates here,
            # so we'll settle for when its content was last modified.
            return stat.st_mtime

class Metadata(Component):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.reset()

    def reset(self):
        self.md = {}

    def load(self, path):
        self.md = get_metadata(path)

    def save(self, path):
        # Save metadata to path in above format
        self.md = remove_empties(self.md)
        if 'Name' not in self.md:
            self.md['Name'] = os.path.basename(path)

        # we no longer want to save this information
        if 'rejected' in self.md:
            del self.md['rejected']
        if 'nsubs' in self.md:
            del self.md['nsubs']
        if 'session' in self.md:
            del self.md['session']

        try:
            with open(os.path.join(path, 'info3.json'), 'w') as f:
                json.dump(self.md, f, indent=1)
        except Exception as e:
            Logger.warn(
                'Metadata: Problem saving info3.json to {:} ({:})'.format(path, e)
            )

    def set(self, field, value=None):
        # set one or more fields of metadata
        if value is None and type(field) == dict:
            for f, v in field.items():
                self.md[f] = v
        else:
            self.md[field] = value

    def get(self, field, default=None):
        # get one or more fields of metadata
        if type(field) == list or type(field) == set:
            d = {}
            for f in field:
                if f in self.md:
                    d[f] = self.md[f]
            return d
        else:
            return self.md.get(field, default)

    def has_changed(self, md):
        # compare md and self.md
        for k, v in self.md.items():
            if k not in md:
                return True
            if md[k] != v:
                return True
        return False



from setuptools import setup, find_packages
from jocular import __version__

import pathlib

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / 'README.md').read_text(encoding='utf-8')

URL = 'https://transpy.eu.pythonanywhere.com/jocular/'

setup(
    name='jocular',
    version=__version__,
    description='A tool for near-live observation of astronomical objects',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=URL,
    author='Martin Cooke',
    author_email='martin.cooke.1917@gmail.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    keywords='astronomy',
    packages=find_packages(exclude=['tests', 'docs']),
    include_package_data=True,
    python_requires='>=3.6',
    install_requires=[
        'numpy>=1.20.1',
        'scikit-image>=0.18.1',  # numerous image ops
        'Click>=7.1.2',  # command line interface
        'kivy>=2.0.0',  # main GUI framework
        'astropy>=4.2',  # fits etc
        'mss>=6.1.0',  #  snapshots
        'pyusb>=1.1.1',  # camera
        'hidapi>=0.10.1',  # filterwheel
        'colour-demosaicing>=0.1.6', # debayering,
        'zwoasi' # wrapper for ZWO cameras 
    ],
    entry_points='''
        [console_scripts]
        jocular=jocular.startjocular:startjocular
        ''',
    project_urls={
        'Bug Reports': 'https://github.com/MartinCooke/jocular/issues',
        'Source': 'https://github.com/MartinCooke/jocular',
    },
)



''' Main app, called after some checks by startjocular
'''

import os
import json
import sys
from pathlib import Path

from kivy.app import App
from kivy.logger import Logger
from kivy.properties import (
    ListProperty,
    NumericProperty,
    ConfigParserProperty,
    OptionProperty,
    StringProperty,
)
from kivy.uix.settings import SettingsWithSidebar
from kivy.clock import Clock
from kivy.core.window import Window
from kivy.core.text import LabelBase

from jocular import __version__

from jocular.component import Component
from jocular.gui import GUI


class Jocular(App):

    showing = OptionProperty(
        'main', options=['main', 'observing list', 'calibration', 'observations']
    )

    highlight_color = ListProperty([1, 1, 1, 1])
    lowlight_color = ListProperty([0.5, 0.5, 0.5, 1])
    hint_color = ListProperty([0.25, 0.25, 0.25, 1])
    lever_color = ListProperty([0.32, 0.32, 0.32, 1])
    background_color = ListProperty([0.06, 0.06, 0.06, 0])  # was 1 transp
    line_color = ListProperty([0.3, 0.3, 0.3, 1])  # used in Obj and Table only

    data_dir = StringProperty(None)

    # quick way to ensure there is a Startup section
    dummy_startup = ConfigParserProperty('dummy', 'Startup', 'dummy', 'app')

    brightness = NumericProperty(1)
    transparency = NumericProperty(0)

    color_codes = {
        'white': [0.8, 0.8, 0.79, 1],
        'red': [0.74, 0.25, 0.3, 1],
        'blue': [0.57, 0.66, 0.82, 1],
        'green': [0.53, 0.69, 0.29, 1],
        'yellow': [0.94, 0.76, 0.31, 1],
    }

    def get_application_config(self):
        return super(Jocular, self).get_application_config(
            os.path.join(self.data_dir, '%(appname)s.ini')
        )

    def get_path(self, name):
        # centralised way to handle accessing resources Jocular needs

        #  if path, return path to data dir, else return path to resource
        if name in [
            'captures',
            'calibration',
            'snapshots',
            'watched',
            'deleted',
            'exports',
            'catalogues',
        ]:
            path = os.path.join(self.data_dir, name)
            try:
                if not os.path.exists(path):
                    Logger.debug('Jocular: creating path {:} -> {:}'.format(name, path))
                    os.mkdir(path)
                return os.path.join(self.data_dir, name)
            except Exception as e:
                sys.exit('Cannot create path {:} ({:})'.format(path, e))
        elif name == 'dsos':
            return os.path.join(self.directory, 'dsos')
        elif name in [
            'observing_list.json',
            'observing_notes.json',
            'last_session.json',
            'capture_scripts.json',
            'previous_observations.json'
        ]:
            return os.path.join(self.data_dir, name)
        elif name == 'shipped_capture_scripts.json':
            return os.path.join(self.directory, 'resources', 'capture_scripts.json')
        elif name == 'libusb':
            return os.path.join(self.directory, 'resources', 'libusb-1.0.dll')
        elif name == 'star_db':
            return os.path.join(self.data_dir, 'platesolving', 'star_tiles.npz')
        elif name == 'dso_db':
            return os.path.join(self.data_dir, 'platesolving', 'dso_tiles')
        else:
            return os.path.join(self.directory, 'resources', name)

    def on_brightness(self, *args):
        for c in [
            'hint_color',
            'highlight_color',
            'lowlight_color',
            'line_color',
            'lever_color',
        ]:
            getattr(self, c)[-1] = self.brightness

    def build_config(self, config):
        # NB this is called before build

        # get user data directory
        try:
            with open(os.path.join(str(Path.home()), '.jocular'), 'r') as f:
                self.data_dir = f.read().strip()
        except Exception as e:
            Logger.exception('Jocular: problem reading user data dir ({:})'.format(e))
            sys.exit()

        self.use_kivy_settings = False

        try:
            with open(self.get_path('configurables.json'), 'r') as f:
                self.configurables = json.load(f)
        except Exception as e:
            Logger.exception('Jocular: cannot read configurables.json ({:})'.format(e))
            sys.exit()

        #  extract all defaults and delete them so settings panel works
        defaults = {}
        for v in self.configurables.values():
            for l in v:
                if ('section' in l) and ('default' in l):
                    sec = l['section']
                    if sec in defaults:
                        defaults[sec][l['key']] = l['default']
                    else:
                        defaults[sec] = {l['key']: l['default']}
                    del l['default']
                if 'component' in l:
                    del l['component']

        for k, v in defaults.items():
            config.setdefaults(k, v)

    def build_settings(self, settings):
        for panel, subsettings in self.configurables.items():
            settings.add_json_panel(panel, self.config, data=json.dumps(subsettings))

    def on_config_change(self, config, section, key, val):
        self.set_param(section, key, val)

    # initially we will call this with all params to handle settings
    def set_param(self, section, key, val=None):

        if val == None:
            val = self.config.get(section, key)

        if section == 'Font sizes':
            setattr(self, '{:}_font_size'.format(key), '{:}sp'.format(val))

        elif section == 'Colours and graylevels':
            if key in ['lowlight_color', 'lever_color']:
                lg = max(30, min(int(val), 100)) / 100
                setattr(self, key, [lg, lg, lg, 1])
            elif key == 'highlight_color':
                self.highlight_color = self.color_codes.get(val, 'white')

        elif section == 'Geometry':
            Window.size = (
                int(self.config.get(section, 'initial_width')),
                int(self.config.get(section, 'initial_height')),
            )

        elif section == 'Filters':
            Component.get('FilterWheel').update_filter(key, val)

    def build(self):


        self.title = 'Jocular v{:}'.format(__version__)

        LabelBase.register(name='Jocular', fn_regular=self.get_path('jocular4.ttf'))

        self.settings_cls = SettingsWithSidebar

        #  apply settings
        for fs in ['ring', 'info', 'form']:
            self.set_param('Font sizes', fs)

        for fs in ['lowlight_color', 'lever_color', 'highlight_color']:
            self.set_param('Colours and graylevels', fs)

        self.set_param('Geometry', 'initial_width')

        # generate GUI so that it exists before we add components
        self.gui = GUI(self.config)

        # draw GUI
        Clock.schedule_once(self.gui.draw, -1)

        return self.gui

    def on_stop(self):
        Component.close()
        self.gui.on_close()
        self.config.write()

    # reset showing to main when any table is hidden
    def table_hiding(self, *args):
        self.showing = 'main'



''' Root widget of the application which handles drawing itself (rings) 
    and its components. Idea is to separate the main gui out entirely. Most 
    elements are defined declaratively in gui.json in the resources dir.
'''

import os
import json
import time
from functools import partial
from collections import OrderedDict

from kivy.app import App
from kivy.uix.floatlayout import FloatLayout
from kivy.clock import Clock
from kivy.logger import Logger
from kivy.core.window import Window
from kivy.properties import BooleanProperty

from jocular.widgets import (JLabel, JToggleButton, JLever, JIconButton,
    JButton, JMulti, JFilterButton)
from jocular.utils import angle360
from jocular.component import Component
from jocular.metrics import Metrics
from jocular.ring import Ring
from jocular.reticle import Reticle
from jocular.widgets import jicon


def _true(val):
    if type(val) == bool:
        return val
    if type(val) == int:
        return val == 1
    if type(val) == str:
        return val.lower() in ['1', 'true', 'yes']
    return False

class GUI(FloatLayout):

    something_has_changed = BooleanProperty(False)

    def __init__(self, config, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        self.disabled_controls = set({})
        self.config = config
        self.gui = OrderedDict()
        self.bind(size=self.redraw)

    def draw(self, dt=None):

        t0 = time.time()

        # construct interface for the first time; we redraw in the order originally drawn
        orig_gui = OrderedDict()
        Component.get('View')

        # handle rings first due to draw order
        origin = Metrics.get('origin')
        radii = Metrics.get('ring_radius')
        thickness = Metrics.get('ring_thickness')

        # location of capture 'saddle'
        angle = [-12.3, 12.2]
        min_a, max_a = angle360(90 - angle[0]), angle360(90 - angle[1])
        greys = {'background': .02, 'outer': .2, 'middle': .17, 'inner': .12, 'image': 0, 'capture_ring': .2}
        start_angle = {'capture_ring': angle360(min(min_a, max_a))}
        end_angle = {'capture_ring': angle360(max(min_a, max_a))}
        for nm, r in radii.items():
            if nm != 'image':
                orig_gui[nm] = {
                    'control_type': 'Ring',
                    'widget': Ring(pos=origin, radius=r, thickness=thickness[nm], grey=greys[nm],
                        start_angle=start_angle.get(nm, 0), end_angle=end_angle.get(nm, 360))
                    } 
 
        # focus/alignment reticle
        orig_gui['reticle'] = {
            'control_type': 'Reticle',
            'widget': Reticle(pos=origin, radius=radii['image'])
        }
 
        # load GUI spec and convert so widget names are keys
        with open(self.app.get_path('gui.json'), 'r') as f:
            gs = json.load(f)

        for comp, widgets in gs.items():
            for name, widget_spec in widgets.items():
                orig_gui[name] = widget_spec
                orig_gui[name]['component'] = comp

        self.gui = OrderedDict()
        # set up a variable to use for new elements such as ring annotations which need a unique name
        self.new_element_count = 0
        for name, spec in orig_gui.items():
            self.draw_element(name, spec)

        self.constructed = True

        # check if the data path is defined and if so, if it is writeable
        data_dir = self.app.data_dir
        try:
            if len(data_dir) > 0:
                with open(os.path.join(data_dir, '.writeable'), 'w') as f:
                    f.write('ok')
                os.remove(os.path.join(data_dir, '.writeable'))
            else:
                raise
        except:
            path = os.path.split(self.app.directory)[0]
            self.app.data_dir = os.path.abspath(os.path.join(path))

        Clock.schedule_once(self.load_components, 0)
        Logger.info('GUI: drawn in {:.0f}ms'.format(1000 * (time.time() - t0)))
           

    def load_components(self, dt):

        t0 = time.time()

        for c in ['Status', 'Metadata', 'Catalogues', 'DSO', 'Notes', 'Session', 'Watcher', 'Stacker', 
            'Capture', 'CaptureScript', 'Observations', 'ObservingList', 'Monochrome',
            'MultiSpectral', 'ObjectIO', 'FilterWheel', 'Aligner', 
            'BadPixelMap', 'Calibrator', 'Snapshotter', 'PlateSolver', 'Annotator']:
            Component.get(c)

        # bind status to components
        Component.bind_status()

        Logger.info('GUI: components loaded in {:.0f}ms'.format(1000 * (time.time() - t0)))
        t0 = time.time()

        # prepare for new object
        Component.initialise_new_object()
        Logger.info('GUI: new object initialised in {:.0f}ms'.format(1000 * (time.time() - t0)))


    def redraw(self, *args):
        ''' Redraw interface by relocating components
        '''

        # if still being constructed, do nothing
        if not hasattr(self, 'constructed'):
            return

        t0 = time.time()
        origin = Metrics.get('origin')
        ring_radius = Metrics.get('ring_radius')
        thickness = Metrics.get('ring_thickness')
 
        for name, spec in self.gui.items():
            if 'widget' in spec:
                w = spec['widget']
                if spec['control_type'] == 'Ring' and name in ring_radius:
                    w.relocate(origin=origin, radius=ring_radius[name], thickness=thickness[name])
                elif spec['control_type'] == 'Reticle':
                    w.pos = origin
                    w.radius = ring_radius['image']
                else:
                    w.relocate(origin=origin, radius=Metrics.get(spec['location']))

        Logger.debug('GUI: redrawn to {:} in {:.0f}ms'.format(Window.size, 1000 * (time.time() - t0)))

    def draw_element(self, wname, spec):
        ''' Draw element and add to gui dictionary for later redrawing
        '''

        control_type = spec['control_type']

        if control_type in {'Ring', 'Reticle'}:
            self.add_widget(spec['widget'])
            self.gui[wname] = spec 
            return

        angle = spec['angle']
        location = spec['location']
        radial = spec.get('radial', False)

        # get initial value from config, else from initial, else make it up

        if self.config.has_option('Startup', wname):
            initial_value = self.config.get('Startup', wname)
        elif 'initial' in spec:
            initial_value = spec['initial']
        elif 'values' in spec:
            initial_value = spec['values'][0]
        else:
            initial_value = 0

        rad = Metrics.get(location)
        orig = Metrics.get('origin')
        radial = spec.get('radial', False)

        # handle icons uniformly
        if 'icon' in spec and 'text' in spec:
            text = jicon(spec['icon'], font_size=spec.get('font_size', None)) + '\n' + spec['text']            
        elif 'icon' in spec:
            text = jicon(spec['icon'], font_size=spec.get('font_size', None))
        else:
            text = spec.get('text', '')

        if control_type == 'JToggleButton':
            w = JToggleButton(angle=angle, origin=orig, radius=rad, text=text, radial=radial,
                allow_no_selection=False)
            if 'group' in spec:
                w.group = spec['group']
            w.state = 'down' if _true(initial_value) else 'normal'

        elif control_type == 'JButton':
            w = JButton(angle=angle, origin=orig, radius=rad, text=text, radial=radial)

        elif control_type == 'JFilterButton':
            w = JFilterButton(angle=angle, origin=orig, radius=rad, text=text, radial=radial)

        # to do: see if we can combine with JButton
        elif control_type == 'JIconButton':
            w = JIconButton(angle=angle, origin=orig, radius=rad, icon=text)
 
        elif control_type == 'JLabel':
            w = JLabel(angle=angle, origin=orig, radius=rad, text=spec.get('text', ''))

        elif control_type == 'JMulti':
            w = JMulti(angle=angle, origin=orig, radius=rad, values=spec['values'],
                text=str(initial_value))

        elif control_type == 'JLever':

            if spec.get('groove', False):
                min_a, max_a = angle360(90 - angle[0]), angle360(90 - angle[1])
                w = Ring(pos=orig, radius=rad, thickness=Metrics.get('outer_width')/2,
                    grey=.13, start_angle=angle360(min(min_a, max_a)), 
                    end_angle=angle360(max(min_a, max_a)))
                #self.gui[_name()] = {
                self.gui['_new{:}'.format(self.new_element_count)] = {
                    'control_type': 'Ring', 
                    'widget': w,
                    'save': False,
                    'location': location}
                self.new_element_count += 1
                self.add_widget(w)

            w = JLever(angles=angle, origin=orig, radius=rad, value=initial_value, 
                values=spec['values'], 
                text=spec.get('text', jicon('lever', font_size=19)),
                radial=spec.get('radial', True),
                springy=spec.get('springy', False))

            self.add_widget(w)  # added here to get depth order right

            if spec.get('major', False):
                for vstr, v in spec['major'].items():
                    lab = JLabel(
                        radial=spec.get('radial', True), text=vstr, 
                        angle=w.value_to_angle(v), origin=orig, radius=rad)
                    self.add_widget(lab) 
                    #self.gui[_name()] = {
                    self.gui['_new{:}'.format(self.new_element_count)] = {
                        'control_type': 'JLabel', 
                        'widget': lab,
                        'save': False,
                        'location': location}
                    self.new_element_count += 1

        # post inits (may need more)
        if not 'icon' in spec:
            for f in ['font_size', 'color', 'text_width', 'background_color', 'halign', 'radial']:
                if f in spec:
                    setattr(w, f, spec[f])
  
        if control_type == 'JLever':
            w.bind(value=partial(self.on_action, wname))
        else:
            self.add_widget(w)
            w.bind(on_press=partial(self.on_action, wname))

        spec['widget'] = w
        self.gui[wname] = spec

    def initialise_component(self, component):
        ''' Push initial settings into the component when it is loaded
        '''

        for k, spec in self.gui.items():
            if spec.get('component', None) == component:
                # do NOT do actions!
                if 'action' not in spec:
                    if 'group' in spec:
                        if spec['widget'].state == 'down':
                            self.on_action(k)
                    else:
                        self.on_action(k)


    def set(self, name, value, update_property=False):
        ''' Called by any component that needs to access GUI element, usually
            for something simple like changing text. Option to pass this
            change back to component via update property/
        '''

        if name not in self.gui:
            return
        spec = self.gui[name]
        ctype = spec['control_type']
        if ctype in ['JLabel', 'JMulti', 'JButton', 'JIconButton']:
            setattr(spec['widget'], 'text', str(value))
        elif ctype == 'JLever':
            getattr(spec['widget'], 'reset_value')(value)
        elif ctype == 'JToggleButton':
            # if it is a group, set all but this to normal
            if 'group' in spec:
                gname = spec['group']
                for n, v in self.gui.items():
                    if v.get('group','') == gname:
                        v['widget'].state = 'down' if n == name else 'normal'
            else:
                setattr(spec['widget'], 'state', 'down' if value else 'normal')
        if update_property:
            self.on_action(name)

    def set_prop(self, name, prop, value):
        if name not in self.gui:
            return
        spec = self.gui[name]
        if prop == 'text' and spec['control_type'] == 'JLever':
            spec['widget'].text = value
        else:
            if hasattr(spec, prop):
                setattr(spec, prop, value)


    def get(self, name):
        ''' Access value of a given component; occasionally required
        '''

        if name not in self.gui:
            return None
        spec = self.gui[name]
        w = spec['widget']
        ctype = spec['control_type']
        if ctype in ['JLabel', 'JMulti']:
            return w.text
        if ctype == 'JLever':
            return w.value
        if ctype == 'JToggleButton':
            return 1 if w.state == 'down' else 0
        return None

    def on_action(self, name, *args):
        ''' Intercept any GUI event in order to load component, before
            passing action on to component via a property change
        '''

        spec = self.gui[name]
        comp = spec['component']
        c = Component.get(comp)  # loads if needed

        if name in self.disabled_controls:
            return

        # if action is specified, simply call that method on component
        if 'action' in spec:
            getattr(c, spec['action'])()
            return

        # otherwise it involves a value change
        control = spec['control_type']

        # handle button groups
        if control == 'JToggleButton':
            if 'group' in spec:
                setattr(c, spec['group'], spec['widget'].text)
                return

        if hasattr(c, name):
            if control == 'JToggleButton':
                setattr(c, name, spec['widget'].state == 'down')
            elif control == 'JLever':
                setattr(c, name, spec['widget'].value)
            elif control == 'JMulti':
                setattr(c, name, spec['widget'].text)

    def enable(self, names=None):
        # None indicates enable everything
        if names is None:
            names = self.disabled_controls
        # remove names from disabled list
        self.disabled_controls = self.disabled_controls - set(names)
        for n in names:
            if n in self.gui:
                spec = self.gui[n]
                spec['widget'].disabled = False

    def disable(self, names):
        # adds names to disabled list
        self.disabled_controls = self.disabled_controls.union(set(names))
        for n in names:
            if n in self.gui:
                spec = self.gui[n]
                spec['widget'].disabled = True

    def on_close(self, *args):
        ''' Save certain GUI values to config. We save everything except
            actions and elements whose save property is False
        '''

        for k, spec in self.gui.items():
            if spec.get('save', True) and 'action' not in spec:
                v = self.get(k)
                if v is not None:
                    self.config.set('Startup', k, v)

    def on_something_has_changed(self, *args):
        # first check if we have an empty stack; if so, nothing has changed
        if self.something_has_changed:
            self.disable(['load_previous', 'quit'])
            spec = self.gui['new_DSO']
            if spec['widget'].text.endswith('new'):
                spec['widget'].text = spec['widget'].text[:-3] + 'save'
        else:
            self.enable(['load_previous', 'quit'])
            spec = self.gui['new_DSO']
            if spec['widget'].text.endswith('save'):
                spec['widget'].text = spec['widget'].text[:-4] + 'new'

